/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2008 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

#ifndef ACCOUNT_SETTINGS_DIALOG
#define ACCOUNT_SETTINGS_DIALOG

#include <akndialog.h> 
#include <aknlists.h>
#include <aknsettingitemlist.h> 

#include "base/util/StringBuffer.h"
#include "CommonOptionsSettingList.h"
#include "ProgressDialog.h"

using namespace Funambol;

// FORWARD DECLARATIONS
class CFunambolAppUi;


/**
 * The account settings dialog, to login the user.
 * If input user/pass is valid, a config sync is started to authenticate the user.
 */
class CAccountSettingsDialog : public CAknDialog
{
private:
    StringBuffer                iUsername;
    StringBuffer                iPassword;
    StringBuffer                iSyncURL;
    
    CFunambolAppUi*             iAppUI;
    TBool&                      iCancelled;
    
    CCommonOptionsSettingList*  iSettingList;

    /// Set to true if the configuration sync has terminated
    bool                        iSyncConfigEnded;
    
    /// If true, the UI is locked for the credentials check (not used now).
    TBool                       iUiLocked;

    /// Set to true if we shall go to the main screen (credentials checked
    /// successfully)
    bool                        iGoToMainScreen;

    /// The progress dialog
    ProgressDialog*             iProgressDlg;
    
    // Counter to enable syncUrl (shortcut).
    int                         iCounterServerUrl;
    static int                  iEnableServerUrl;

private:
    
    CAccountSettingsDialog(CFunambolAppUi* appUi, TBool& aCancelled);
    ~CAccountSettingsDialog();
    
    /// Overrided to create the setting list at runtime. 
    void PreLayoutDynInitL();
    
    /// Overrided to handle dialog's menu commands.
    void ProcessCommandL(TInt aCommand);
    
    /// Overrided to capture the key events: up/down arrow, selection key.
    TKeyResponse OfferKeyEventL(const TKeyEvent &aKeyEvent, TEventCode aType);
    
    /// Invoked before leaving the dialog. Decides if we need to save or cancel.
    TBool OkToExitL(TInt aButtonId);
    
    /**
     * Loads account data from client configuration into the UI dialog.
     */
    void  load();
    
    /**
     * Saves account data from UI dialog into the client configuration.
     * If data is not valid or unchanged, exits with EFalse.
     * @return  ETrue if settings saved correctly, EFalse if not.
     */
    TBool save();
    
    /**
     * Used to lock/unlock the UI during config sync. 
     * A progress dialog is prompted, and cba buttons are disabled.
     */
    void lockUI(const TBool aEnable);
    
    /// Saves the old account settings data into the configuration.
    /// This method is called when new credentials are wrong and the user cancelled the login.
    void restorePreviousSettings();

    void checkToEnableServerUrl(const TKeyEvent &aKeyEvent, TEventCode aType);
    
    /**
     * Checks the input syncURL and adds default port (80) if missing.
     * For example: "http://<hostName>/sync" -> "http://<hostName>:80/sync"
     * If the port number is already defined, nothing is done.
     * 
     * @note This is necessary for FP1 devices, where the port number must be explicitly set in
     * SyncML profile, otherwise it won't work.
     * 
     * @param syncUrl  [IN-OUT] the synchronization URL to check
     * @return         true if the port number is added, false if nothing done
     */
    bool addPortIfMissing(StringBuffer& syncUrl);
     
public:
    
    /**
     * Creates a new CAccountSettingsDialog object.
     * Calls the CAknDialog::ConstructL, passing the menu resource.
     * @param aMenuTitleResourceId  the resource ID of the menubar for this dialog
     * @param aAppUI                pointer to the FunambolAppUI object that called this method
     * @param aCancelled            [OUT] is set to 'true' if the user cancelled the login
     */
    static CAccountSettingsDialog* NewL(TInt aMenuTitleResourceId, 
                                        CFunambolAppUi* appUi, 
                                        TBool& aCancelled);

    /**
     * This is a callback invoked when the configuration sync ended
     * @param aGoToMainScreen this flag indicates that the credentials have been
     * verified and we are ready to go the main screen
     */
    void CheckCredentialsEnded(bool aGoToMainScreen);


};

#endif  // ACCOUNT_SETTINGS_DIALOG

