/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2008 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

#ifndef CLIENT_SETTINGS_INC
#define CLIENT_SETTINGS_INC

#include "base/util/StringBuffer.h"
#include "client/DMTClientConfig.h"
#include "Funambol.hrh"
#include "commonDefs.h"
#ifdef __GCCE__ 
#include "SyncEngine.h"
#else
#include "FakeSyncEngine.h"
#endif

using namespace Funambol;

class ClientSettings : public DMTClientConfig
{
public:
    static ClientSettings* getInstance();

    virtual bool save();
    virtual bool read();

    /**
     * Loads the ConfigSyncSource settings (meaning: handled by the config sync source)
     * @return  true if no errors
     */
    bool readConfigSyncSourceSettings();

    /**
     * Called the first time the client runs.
     * The config is not found, so we generate it at runtime.
     */
    void createDefaultConfig();
    
    /**
     * Called every time the Client starts, to check if a new
     * version has been installed. If true, we execute proper actions
     * here to upgrade the client configuration.
     * When installing a new version, we only copy the file '\upgrade\config.ini'
     * with all necessary information - here we check that file to really
     * upgrade the client at runtime.
     * 
     * @return true if the configuration has been upgraded to a new version
     */
    bool makeUpgradeActions();
    
    /**
     * Creates the SyncML profile with name 'SYNC_PROFILE_NAME' and
     * propagates the current configuration settings into the profile.
     * @return  true if profile created without errors
     */
    bool createSyncMLProfile();
    
    /**
     * Creates the IMAP account with name 'SYNC_PROFILE_NAME' and
     * propagates the current configuration settings into the account.
     * If the account already exists, it just updates the account settings
     * and returns false.
     * @return  true if account created without errors
     */
    bool createIMAPAccount();
    
    /**
     * Utility method to check if credentials are empty.
     * We check username, password and syncURL values.
     * @return true if at least one credential field is empty.
     */
    bool isCredentialsEmpty();
    
    /**
     * Checks if we copied the account data from existing SyncML profile.
     * Usually false, it may happen if it's the first execution of the Client and
     * a profile named SYNC_PROFILE_NAME is already existing (since 7.1.12).
     */
    bool checkCopiedAccountFromProfile();

    
    // ---- get methods ----
    const StringBuffer& getDisplayName()       { return displayName;       }
    const StringBuffer& getEmailAddress()      { return emailAddress;      }
    const TSyncUsing    getSyncUsing()         { return syncUsing;         }
    const bool          getPendingSyncConfig() { return pendingSyncConfig; }
    const TInt          getSchedulerMinutes()  { return schedulerMinutes;  }
    const TInt          getCpInterval()        { return cpInterval;        }
    const StringBuffer& getIapName()           { return iapName;           }
    const bool          getFirstRun()          { return firstRun;          }
    const TInt64        getLastEmailSyncTimestamp() { return lastEmailSyncTimestamp; }
    const TLastSyncStatus getLastEmailSyncStatus()    { return lastEmailSyncStatus; }

    // ---- set methods ----
    void setPendingSyncConfig       (const bool aPendingSyncConfig);
    void setSyncUsing               (const TSyncUsing aSyncUsing);
    void setLastEmailSyncTimestamp  (TInt64 aTime);
    void setLastEmailSyncStatus     (TLastSyncStatus aStatus);
    void setSchedulerMinutes        (const TInt aSchedulerMinutes);
    void setCpInterval              (const TInt aCpInterval);
    void setIapName                 (const StringBuffer& aIapName);
    void setImapAccountLuid         (const StringBuffer& aLuid);
    void setFirstRun                (const bool aFirstRun);
    void setPassword(const StringBuffer& aPassword);
    void setSyncURL(const StringBuffer& aSyncURL);
    void setUsername(const StringBuffer& aUsername);

    void forceDirtyOnMainNode();


private:
    
    ClientSettings();
    
    /**
     * Creates and return a StringBuffer with the deviceID: "IMEI:xxxxxxxx".
     */
    StringBuffer retrieveDeviceId();
    
    StringBuffer getImeiL();

    /**
     * Returns a property value from its name, reading from the upgrade dir.
     * Upgrade dir is under 'APPLICATION_URI/upgrade' path.
     */
    StringBuffer getPropertyFromUpgradeDir(const StringBuffer& propertyName);
    
    /// Formats and return a UserAgent string, based on the sw version.
    /// In the format "<application name> v. x.y.z".
    StringBuffer createUserAgent(const StringBuffer& swv);
    
    // --- Not used: dirs are created automatically by DMNode if fopen fails ---
    //TInt createAllDirs(const StringBuffer& rootDir, RArray<TInt>& sourceIDs);
    //TInt makeDir(const StringBuffer& aPath, const char* aContext, RFs* aFileSession);

    
private:
    static ClientSettings* instance;
    static const StringBuffer configContext;
    static const StringBuffer upgradeContext;
    static const StringBuffer addressNodeName;
    static const StringBuffer displayNodeName;
    static const StringBuffer nodePropertyName;
    static const StringBuffer pendingSyncConfigPropertyName;
    static const StringBuffer syncUsingPropertyName;
    static const StringBuffer lastEmailSyncTimestampPropertyName;
    static const StringBuffer lastEmailSyncStatusPropertyName;
    static const StringBuffer schedulerMinutesPropertyName;
    static const StringBuffer cpIntervalPropertyName;
    static const StringBuffer iapNamePropertyName;
    static const StringBuffer firstRunPropertyName;

    StringBuffer    emailAddress;
    StringBuffer    displayName;
    bool            pendingSyncConfig;
    TSyncUsing      syncUsing;
    TInt64          lastEmailSyncTimestamp;
    TLastSyncStatus lastEmailSyncStatus;
    TInt            schedulerMinutes;
    TInt            cpInterval;        
    StringBuffer    iapName;
    bool            firstRun;      // It's set to true when the config is first generated.
    bool            dirty;         // Is the main node dirty?
    bool            nativeDirty;   // Are native profiles dirty?
    bool            copiedAccountFromProfile;       // If we copied the account data from existing SyncML profile
};

#endif

