/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2008 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */


#ifndef SYNCENGINE_H_
#define SYNCENGINE_H_

#include <e32base.h>
#include <e32std.h>
#include <syncmlclient.h>
#include <syncmlclientds.h>
#include <syncmldef.h>
#include <syncmlobservers.h>
#include "base/fscapi.h"
#include "base/util/stringUtils.h"
#include "Funambol.hrh"

#ifdef FAKE_IMAP_ACCOUNT
#include "FakeImapEmailEngine.h"
#else
#include "ImapEmailEngine.h"
#endif



// FORWARD DECLARATIONS
class CFunambolAppUi;
class CSyncScheduler;

using namespace Funambol;


/** 
 * Synchronization status of current sync.
 */
enum TCurrentSyncStatus
{
    ESyncStatusConnecting,
    ESyncStatusConnected,
    ESyncStatusDisconnected,
    ESyncStatusLoggingOn,
    ESyncStatusLoggedOn,
    ESyncStatusSendingData,
    ESyncStatusReceivingData,
    ESyncStatusMappings,
    ESyncStatusCompleted,
    ESyncStatusNotCompleted,
    ESyncStatusSendingItemNumber,
    ESyncStatusReceivingItemNumber,
    ESyncStatusError,
    ESyncStatusCanceled
};

/**
 *  Possible status of last sync.
 */
enum TLastSyncStatus
{
    ELastSyncSuccessful,
    ELastSyncFailed,
    ELastSyncNotFound
};


/**
 *  The sync can be triggered only from these runners.
 */
enum TSyncRunners
{
    ESyncTriggeredManual,           // From UI buttons
    ESyncTriggeredPush,
    ESyncTriggeredClientPush,
    ESyncTriggeredAsyncAction,      // From Active object's RunL()
    ESyncTriggeredScheduled
};


/**
 * Struct used to pair a jobID with a sourceID, used to understand 
 * which source is synced given the current job ID.
 * Also used to flag a job as 'to be aborted' once it starts.
 */
typedef struct JobEntry {
    TSmlJobId   jobID;        /**< The job ID */
    TInt        sourceID;     /**< The corresponding source ID */
    TBool       toAbort;      /**< Set to 'ETrue' if you want to abort the job once it starts */
} JobEntry;



/**
 * This class is responsible of the whole synchronization process.
 * Use StartSyncL() to start a new sync session, StopL() to stop it, 
 * CheckSyncInProgress() to know if the sync is running.
 * This class is registered to receive SyncML notifications during the sync,
 * so they will be forwarded to the parent CFunambolAppUi.
 * CSyncEngine extends CActive, to implement an active object.
 * Push triggered syncs arrives from a different thread, but we can execute them 
 * only from the original thread that owns the RSyncMLSession, so we use the
 * active object's RunL() to run asynchronous actions within the right thread.
 */
class CSyncEngine : public CActive, 
                    public MSyncMLProgressObserver, public MSyncMLEventObserver
{

private: 
    
    /// Pointer to the object that created this instance, the appUI.
    /// Used to callback methods on sync events notifications.
    CFunambolAppUi*             iAppUI;
     
    RSyncMLSession              iSyncMLSession;     
    TSmlProfileId               iActiveProfileID;

    /**
     * Used for scheduled (and delayed) synchronizations.
     */
    CSyncScheduler*              iScheduler;

    /// The current source that is under sync.
    TInt                        iCurrentSource;

    ///The current syncML status: one of TCurrentSyncStatus.
    TCurrentSyncStatus          iCurrentSyncStatus;
 
    
    //TErrorLevel                 iLastErrorLevel;
    //TInt                        iLastErrorCode;
    //RBuf                      iLastErrorMsg;
    
    /// Counter of total items received (Server -> Client) for this job.
    /// It's resetted when the job starts (see ManageNewJobStarted())
    TInt                        iTotalClientItems;
    
    /// Counter of total items sent (Client -> Server) for this job.
    /// It's resetted when the job starts (see ManageNewJobStarted())
    TInt                        iTotalServerItems;

    /**
     * List of queued sync jobs.
     * It's necessary to retrieve the sourceID once the job starts, given
     * the job ID: otherwise we cannot know which source is under sync.
     * A new entry is appended every time a job is created, and it's
     * removed once the job really starts (it's no more useful).
     * It's also used to flag a queued job that need to be stopped, so
     * it will be aborted when the job really starts.
     */
    RArray<JobEntry>            iQueuedJobs;

    /// Used to avoid manipulating the iQueuedJobs 
    /// member between different threads.
    RCriticalSection            iCriticalQueuedJobs;

    /**
     * Enumeration of asynchronous SyncML actions.
     * These are actions that can be executed asynchronously, using
     * the active object RunL().
     */
    enum TAsyncAction
    {
        EIdle,
        EStartSync,
        EStopSync,
        EEnableScheduler
    };
    
    /**
     * The requested asynchronous SyncML action.
     * The active object will check this attribute to know
     * which action need to be executed, one of TAsyncAction.
     */
    TAsyncAction               iAsyncAction;
    
    /**
     * A list of source IDs, requested to sync
     * through the active object.
     */
    RArray<TInt>                iRequestedSourceList;

    /**
     * The ID of the thread that owns the RSyncMLSession.
     * @note  It's mandatory to originate and complete an asynchronous 
     *        request inside the same thread, so here is the ID of the sole
     *        thread used to originate and complete requests.
     */
    TInt                        iMainThreadID;

    /// Imap email engine (for email synchronization)
    CImapEmailEngine*           emailEngine;

    /** Number of sources that are under synchronization and have not
     * terminated. This value is initialized at the beginning of a sync
     * and updated as each source terminates.
     */
    TInt iNumSources;
    
    /// Counter used when the SyncML profile is not found (see getActiveProfileID())
    /// Note: internal use, do not modify its value!
    TInt                        iCreateProfileRetry;

   
private:

    /// Default constructor
    CSyncEngine();

    /**
     * Initialize members, adds this object to the Active Scheduler.
     * Calls InitializeSession() to init SyncML session.
     * @param aAppUI  the parent CFunambolAppUi, used for callbacks
     */
    void ConstructL(CFunambolAppUi* aAppUI);

    /**
     * Initializes the SyncML session, opening RSyncMLSession object.
     * Register this class to receive SyncML events notifications and
     * loads the default profile.
     * @note  may leave with ERR_PROFILE_NOT_FOUND or other Symbian error codes
     */
    void InitializeSessionL();
    

public:
    /**
     * Starts a new sync job, only for 1 desired source.
     * Gets the corresponding task ID from the source ID, and creates a new
     * sync job with only this task.
     * If another sync is currently running, this job is automatically
     * queued by the Native Sync Client - so the job could effectively not start
     * right here, it's only triggered to start.
     * 
     * @param aSourceID  the source ID of the desired source (e.g. SOURCE_CONTACTS)
     */
    void StartSync(const TInt aSourceID);

    void syncConfigEnded(TInt status);

private:


    /// Sets the iCurrentSource member
    void SetCurrentSource(const TInt aSourceID) { iCurrentSource = aSourceID; }

    //void setSyncError(const TErrorLevel aErrorLevel, const TInt aError);
    
    
    /**
     * Returns the datasource name for a desired source.
     * It just scans a static struct.
     * @param aSourceID  the source ID       (e.g. SOURCE_CONTACTS)
     * @return           the datasource name (e.g. "C:Contacts.cdb")
     */
    TBuf<32>   GetDataSourceFromSourceID(const TInt aSourceID);
    
    /**
     * Returns the sourceID from the datasource name.
     * It just scans a static struct.
     * @param aDataSource  the datasource name (e.g. "C:Contacts.cdb")
     * @return             the source ID       (e.g. SOURCE_CONTACTS)
     */
    TInt   GetSourceIDFromDataSource(const TDesC& aDataSource);
    
    /**
     * Returns the taskID corresponding to a source ID.
     * The taskID can change on different devices, so we retrieve it
     * scanning all tasks from the active SyncML profile.
     * @param aSourceID  the source ID  (e.g. SOURCE_CONTACTS)
     * @return           the task ID, KErrNotFound if task not found
     */
    TSmlTaskId GetTaskIDFromSourceID(const TInt aSourceID);
    
    /**
     * Returns the sourceID corresponding to a taskID.
     * The taskID can change on different devices, so we retrieve it
     * scanning all tasks from the active SyncML profile.
     * @param aTaskID  the task ID
     * @return         the source ID (SOURCE_CONTACTS if not found)
     */
    TInt GetSourceIDFromTaskID(const TSmlTaskId aTaskID);

    /**
     * Checks if a sync job is currently running, and returns its ID.
     * @retun  the job ID of the running sync 
     *         '-1' if no sync job is running ('0' is a valid job ID!)
     */
    TSmlJobId GetCurrentJobID();
    
    /**
     * Called when a new sync job effectively starts, it handles
     * the beginning of a new source synced. 
     * Sets the 'iCurrentSource' attribute, to the corresponding sourceID.
     * If the job was flagged to be aborted, it's stopped here.
     * @param aJobID  ID of the job jist started
     */
    void ManageNewJobStarted(const TSmlJobId aJobID);
    
    /**
     * Stops the running job, by its ID.
     * @return  0 : no errors
     *         -1 : error opening job
     *         -2 : error stopping job
     */
    TInt StopJob(const TSmlJobId aJobID);
    
    /**
     * Stops all the queued SyncML jobs.
     * Mark the queued jobs with the abort flag.
     * When those jobs will start, they will be stopped immediately.
     */
    void StopQueuedJobs();
    
   
    /**
     * Appends a new JobEntry into the iQueuedJobs array.
     * @param aJobID     the job ID
     * @param aSourceID  the source ID
     * @param aToAbort   [OPTIONAL] if the job need to be aborted once started (default = false)
     * @return           the new number of queued jobs
     */
    TInt AddQueuedJob(const TSmlJobId aJobID, const TInt aSourceID, const TBool aToAbort = EFalse);
    
    /// Returns the JobEntry index inside the iQueuedJobs array, given the job ID.
    TInt GetQueuedJobIndex(const TSmlJobId aJobID);
    
    /// Returns true if the passed source ID is found in the iQueuedJobs list.
    TBool findSourceIDInQueuedJobs(const TInt aSourceID);
    
    /**
     * Returns true if that source ID is currently under sync and in 'receiving' state.
     * It's a workaround for ClientPush, as we need to ignore the CP
     * notifications that happen during a sync, when items are added from
     * the Server. Otherwise, a new sync will be started every time when
     * the current sync is ended.
     */
    TBool isSourceUnderSyncAndReceiving(const TInt aSourceID);

    
    // ----------- Active Object's methods ------------
    /**
     * Used to request an asynchronous action to be executed.
     * Sets the 'iAsyncAction' member and signals that the asynchronous 
     * request associated with 'iStatus' is complete, so that the 
     * active object's RunL() will be called by active scheduler.
     * @note   we must signal the status in the same thread that originated the request.
     * 
     * @param aAction  the action to be executed by the active object, one of TAsyncAction
     */
    void RequestAsyncAction(const TAsyncAction aAction);
    
    /**
     * Handles active object's request completion events.
     * It's called by the active scheduler when a request completion event occurs.
     * The asynchronous action to be executed is defined by 'iAsyncAction' member,
     * which is previously set when the request is raised (in RequestAsyncAction()).
     * 
     * The active object is used to execute asynchronous actions like StartSync()
     * inside the same thread that created the SyncML session - it's used to start
     * synchronizations from push or client push.
     */
    void RunL(void);
    
    /**
     * Implements cancellation of  outstanding request.
     * This function is called as part of the active object's Cancel().
     */
    void DoCancel(void);
    
    /// Called by the active scheduler if the active object's RunL() function leaves.
    TInt RunError(TInt aError);
    // ---------------------------------------
    

public:
    /**
     * perform configuration synchronization. This is not handled by native
     * API and it is performed using Funambol APIs.
     */
    void syncConfig();

private:
 
    /**
     * perform email synchronization. This is not handled by native
     * API and it is performed using IMAP protocol
     */
    TInt syncEmail();


    TLastSyncStatus GetLastPimSyncStatus(const TInt aSourceID,
                                         TTime* aLastSyncTime,
                                         TInt* aLastSyncResult);

    TLastSyncStatus GetLastEmailSyncStatus(TTime* aLastSyncTime,
                                           TInt*  aLastSyncResult);
    
    /**
     * Returns the corresponding syncType enumeration from the text decribing it.
     * In config, syncType is stored as text.
     * For example: "one-way-client" -> ESmlOneWayFromClient.
     */
    TSmlSyncType getSyncTypeFromText(const StringBuffer& text);

	/**
	 * Proper actions when a SyncML error is captured.
	 * Filters errors by type, and calls appUI->DisplayMessage() for some of them.
	 */
    void ManageSyncMLErrors(const TInt aError, const TInt aAdditionalData);


    /**
     * This method checks if we are at the end of a sync, and in that case it
     * starts the UpgradeManager to check if there is a new version available
     */
    void checkForUpdates();
    
    
protected:
    
    /*
     * SyncML ClientAPI Observers
     */
    void OnSyncMLSyncProgress(TStatus aStatus, TInt aInfo1, TInt aInfo2);
    void OnSyncMLSyncError(TErrorLevel aErrorLevel, TInt aError, TInt aTaskId,
                           TInt aInfo1, TInt aInfo2);
    void OnSyncMLDataSyncModifications(TInt aTaskId,
                                       const TSyncMLDataSyncModifications &aClientMod,
                                       const TSyncMLDataSyncModifications &aServerMod);
    
    // From MSyncMLEventObserver
    void OnSyncMLSessionEvent(TEvent aEvent, TInt aIdentifier, TInt aError,
                              TInt aAdditionalData);
    
    
public:

    static CSyncEngine* NewL(CFunambolAppUi* aAppUI);
    static CSyncEngine* NewLC(CFunambolAppUi* aAppUI);

    /// Destructor
    ~CSyncEngine();
    
    
    /**
     * Starts a new sync job, for an array of sources.
     * Will call the StartSyncL(sourceID) for each source: the first
     * First task starts immediately, other tasks are automatically queued.
     * 
     * @param aSourceIDList  the array of source IDs to be synced
     * @param from           defines who triggered the sync action, one of TSyncRunners
     */
    void StartSync(const RArray<TInt> &aSourceIDList, const TSyncRunners from);
    
    
    /**
     * If a sync is running, tries to abort the current sync.
     * @return  1 if current sync stopped 
     *          0 if no job running
     *         <0 in case of errors stopping the current job
     */
    TInt StopSync();

    /**
     * Checks if a sync job is currently running.
     * @retun  Etrue if a sync job is running, EFalse if not.
     */
    TBool CheckSyncInProgress();
    
    
    /**
     *  Returns the source ID currently under sync.
     * It's important to know which is the task running, for example
     * to refresh the correct pane corresponding to it.
     */
    TInt GetCurrentSourceID() { return iCurrentSource; }
    
    
    /**
     * Opens the SyncML history and finds the last sync status for the desired source.
     * If the source was never synchronized, returns ELastSyncNotFound 
     * (aLastSyncTime and aLastSyncResult are not touched). 
     * Otherwise on exit 'aLastSyncTime' param will contain the last timestamp 
     * (in local time) and 'aLastSyncResult' the last sync result code.
     * This method scans the RSyncMLHistoryLog for the active profile, 
     * starting from the most recent entry and searching for one that
     * contains the desired task (corresponding to the passed sourceID).
     * 
     * @param aSourceID         the source ID of the desired sorce to look for
     * @param aLastSyncTime     [OUT] the last sync timestamp, as a TTime object (in local time)
     * @param aLastSyncResult   [OUT] the last sync result code
     * #return                  the last sync status, one of TLastSyncStatus.
     */
    TLastSyncStatus GetLastSyncStatus(const TInt aSourceID, TTime* aLastSyncTime,
                                      TInt* aLastSyncResult);
    
    
    TSyncMode getSourceSyncMode(const TInt aSourceID);
    
    /// Also saves the config.
    void setSourceSyncMode(const TInt aSourceID, const TSyncMode aSyncMode);
    
    /// Also saves the config.
    void setSourceTimestamp(const TInt aSourceID, const TUint timestamp);

    void EmailSyncEnded(TCurrentSyncStatus status);


    
    /**
     * Enables the sync scheduler.
     * Sets the scheduler to load the list of active sources every time itself,
     * then enables the scheduler with the given interval.
     */
    void enableScheduler(const bool aIsRepetitive, const unsigned int aInterval);
    
    /**
     * Disables the sync scheduler (proxy method).
     * If a scheduled sync is already set, it will not complete because
     * it will be discarded on RunL().
     */
    void disableScheduler();
    
    /// Checks the sync scheduler (proxy method).
    bool isSchedulerEnabled();
    
    /// Removes a sourceID from the scheduler list (proxy method).
    bool removeFromScheduler(const TInt aSourceID);
    
    // Proxy method, used by the scheduler.
    RArray<TInt> GetActiveSourcesList();
    
    /**
     * Get the active pim sync profile (-1 if it does not exist)
     */
    TInt getActiveProfileID();
};

#endif  // SYNCENGINE_H_
