/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2008 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

#ifndef SYNC_SETTINGS_DIALOG
#define SYNC_SETTINGS_DIALOG

#include <akndialog.h> 
#include <aknlists.h>

// FORWARD DECLARATIONS
class CFunambolAppUi;

using namespace Funambol;

typedef TBuf<32> SettingsLabel;


/**
 * The sync settings dialog, containing a list of sub-options to choose.
 * Contains a settings listbox as a double line + large icon.
 * First line is created at runtime, from iListboxLabels:
 *    - Emails
 *    - Contacts
 *    - Calendar
 *    - Sync Mode
 * The second line is composed at runtime, reading configuration settings 
 * of each source and sync method.
 */
class CSyncSettingsDialog : public CAknDialog
{

private:
    
    /// Pointer to the object that created this instance, the appUI.
    /// Used to callback methods to the appUI, which is the controller.
    CFunambolAppUi*              iAppUI;
    
    /// The listbox of text labels.
    /// It's linked at startup to the LISTBOX resource declared inside this dialog.
    CAknDoubleLargeStyleListBox* iListBox;
    
    /// The total number of items for the settings listbox.
    TInt                         iListboxNumberOfItems;
    
    /// Array of listbox labels: "Email", "Contacts",...
    RArray<SettingsLabel>        iListboxLabels;
    
    // Position of items in the settings UI. They're initialized in constructor.
    TInt iContactsItemIdx;
    TInt iCalendarItemIdx;
    TInt iNotesItemIdx;
    TInt iSyncMethodItemIdx;
    TInt iLoggingItemIdx;
    
    
    /// Calls the CAknDialog::ConstructL, with the menu resource.
    void ConstructL(TInt aMenuTitleResourceId, CFunambolAppUi* aAppUI);
    
    /// Default constructor
    CSyncSettingsDialog();
    
    /// Overrided to update the listbox labels at runtime. 
    void PreLayoutDynInitL();
    
    void PostLayoutDynInitL();
    
    // Overrided to handle dialog's menu commands.
    // Not used: we don't use a menu!
    //void ProcessCommandL(TInt aCommand);
    
    /// Invoked before leaving the dialog. Decides if we need to save or cancel.
    TBool OkToExitL(TInt aButtonId);

    /// Overrided to capture the key events: up/down arrow, selection key.
    TKeyResponse OfferKeyEventL(const TKeyEvent &aKeyEvent, TEventCode aType);
    
    
    /**
     * This method is called only once during dialog creation.
     * 'iListBox' object is linked to the listbox loaded from resources.
     * Labels are updated with string resources.
     */
    void CreateListBoxL();
    
    /// Shows the sub-option dialog.
    void ShowOptionDialog(const TInt aPane);
    void ShowEmailOptionsDialog();
    void ShowContactsOptionsDialog();
    void ShowCalendarOptionsDialog();
    void ShowNotesOptionsDialog();
    void ShowSyncMethodDialog();
    void ShowLoggingDialog();

    /**
     * Updates all listbox items dynamically, reading resources strings and
     * composing the status of each item, reading the config.
     * This method is called when the dialog is created.
     */
    void RefreshAllListboxItems();
    
    // Update the specific item status on the UI listbox.
    void RefreshEmailStatus();
    void RefreshContactsStatus();
    void RefreshCalendarStatus();
    void RefreshNotesStatus();
    void RefreshSyncMethodStatus();
    void RefreshLoggingStatus();
    
    /// Updates the text of the desired listbox item, given the index.
    void RefreshItemText(const TInt aItemIndex, const StringBuffer& aText);
    void RefreshItemText(const TInt aItemIndex, const TDesC& aText);
    
    /**
     * Compose the status of a desired source (to be displayed in the 2nd line of the listbox).
     * @param aSourceID  ID of the source desired
     * @param status     [OUT] the status string to be formatted
     */
    void composeSourceSettingsStatus(const TInt aSourceID, StringBuffer& status);
    
    /**
     * Compose the status of the syncMethod item (to be displayed in the 2nd line of the listbox).
     * @param status     [OUT] the status string to be formatted
     */
    void composeSyncMethodStatus(StringBuffer& status);
    
    /**
     * Compose the status of the Logging item (to be displayed in the 2nd line of the listbox).
     * @param status     [OUT] the status string to be formatted
     */
    void composeLoggingStatus(StringBuffer& status);
    
    
    /**
     * Returns the resourceID of the settings item label of index 'aItemIndex'.
     * If not found, returns KErrNotFound (-1).
     */
    TInt getSettingsLabelResourceID(const TInt aItemIndex);
    
    
public:
    
    /**
     * Creates a new CSyncSettingsDialog object.
     * @param aMenuTitleResourceId  the resource ID of the menubar for this dialog
     * @param aAppUI                pointer to the FunambolAppUI object that called this method
     */
    static CSyncSettingsDialog* NewL(TInt aMenuTitleResourceId, CFunambolAppUi* aAppUI);
    
};

#endif  // SYNC_SETTINGS_DIALOG

