/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2008 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */


#include <e32base.h>
#include <e32cmn.h>
#include <cntdb.h>

#include "base/Log.h"

#include "FunambolAppUi.h"
#include "CleanupContactsManager.h"

using namespace Funambol;

/**
 * CCleanupContactsManager::NewL()
 * Two-phased constructor.
 */
CCleanupContactsManager* CCleanupContactsManager::NewL(CFunambolAppUi* appUi)
{
    CCleanupContactsManager* self = CCleanupContactsManager::NewLC(appUi);
    CleanupStack::Pop(self);
    return self;
}

/**
 * CCleanupContactsManager::NewLC()
 * Two-phased constructor.
 */
CCleanupContactsManager* CCleanupContactsManager::NewLC(CFunambolAppUi* appUi)
{
    CCleanupContactsManager* self = new(ELeave) CCleanupContactsManager(appUi);
    self->ConstructL();
    CleanupStack::PushL(self);
    return self;
}

/**
 * CCleanupContactsManager::CCleanupContactsManager()
 * C++ default constructor can NOT contain any code, that might leave.
 */
CCleanupContactsManager::CCleanupContactsManager(CFunambolAppUi* appUi)
                                          : CActive(CActive::EPriorityStandard),
                                            iContactItemId(NULL),
                                            iCanceled(EFalse),
                                            appUi(appUi)
{
}

/**
 * CCleanupContactsManager::ConstructL()
 * Symbian 2nd phase constructor can leave.
 */
void CCleanupContactsManager::ConstructL()
{
    LOG.debug("ConstructL");

    // open contacts database
    iContactsDb = CContactDatabase::OpenL();
    // register my observer
    iNotifier = CContactChangeNotifier::NewL(*iContactsDb, this);
    // create iterator for contacts database
    iContactIter = new (ELeave) TContactIter(*iContactsDb);
    
    // adds itself to the scheduler
    CActiveScheduler::Add(this);

    LOG.debug("ConstructL done");
}

/**
 * CCleanupContactsManager::~CCleanupContactsManager()
 * Class destructor
 */
CCleanupContactsManager::~CCleanupContactsManager()
{
    if( IsActive() ) {
        Cancel();
    }

    delete iNotifier;
    iNotifier = NULL;
    delete iContactsDb;
    iContactsDb = NULL;
    delete iContactIter;
    iContactIter = NULL;
}

/**
 * CCleanupContactsManager::DeleteAllEntriesL()
 * Start deleting all contacts stored in the default contacts database.
 * Please note that: "Each item has an access count. An item cannot be fully
 * deleted unless its access count is zero. "
 */
void CCleanupContactsManager::DeleteAllEntriesL()
{
    LOG.info("Cleaning up all Contacts...");

    // obtain first contact id
    iContactItemId = iContactIter->FirstL();
    // start active scheduler activity
    TRequestStatus *s = &iStatus;
    User::RequestComplete(s, KErrNone);

    SetActive();
}

/**
 * CCleanupContactsManager::HandleDatabaseEventL()
 * Implements virtual method of MContactDbObserver.
 */
void CCleanupContactsManager::HandleDatabaseEventL(TContactDbObserverEvent aEvent)
{
    switch (aEvent.iType)
    {
        case EContactDbObserverEventUnknownChanges:
            // TODO All contacts deleted (notify listener)
            break;
        default:
            break;
    }
    
}

void CCleanupContactsManager::Stop(void)
{
    iCanceled = ETrue;
}

void CCleanupContactsManager::DoCancel(void)
{
    iCanceled = ETrue;
}


void CCleanupContactsManager::RunL(void)
{
    if(iCanceled)
    {
        //operation has been stopped
        appUi->contactsCleanupCompleted();
    }
    else if(iContactItemId == KNullContactId)
    {
        // last element in contact database (or empty db)
        // nothing to be deleted
        appUi->contactsCleanupCompleted();
    }
    else
    {
        // operation was not canceled and the item is valid
        TInt err = KErrNone;
        TRAP(err,iContactsDb->DeleteContactL(iContactItemId));
        if(err == KErrDiskFull)
        {
            // unrecoverable error
            // dismiss wait dialog
            // TODO alert user
            // TODO act upon
            LOG.error("Cannot delete contact: disk full");
            appUi->contactsCleanupCompleted();
        }
        else // KErrNone, KErrNotFound, KErrInUse    
        {
            iContactItemId = iContactIter->NextL();
        
            TRequestStatus* status = &iStatus;
            User::RequestComplete(status, KErrNone);
            SetActive();
        }
    }
}





