/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2008 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

#include <Etel3rdParty.h>

#include "base/util/stringUtils.h"
#include "base/util/symbianUtils.h"
#include "client/DMTClientConfig.h"
#include "spdm/DMTree.h"

#include "commonDefs.h"
#include "customization.h"
#include "utils.h"
#ifdef __GCCE__
#include "SyncProfileManager.h"
#else
#include "FakeSyncProfileManager.h"
#endif
#include "ClientSettings.h"
#include "DefaultSymbianConfigFactory.h"
#include "UpdateManager.h"

#ifdef FAKE_IMAP_ACCOUNT
#include "FakeImapAccountManager.h"
#else
#include "ImapAccountManager.h"
#endif

using namespace Funambol;

ClientSettings* ClientSettings::instance = NULL;

const StringBuffer ClientSettings::configContext("/Email");
const StringBuffer ClientSettings::upgradeContext("/upgrade");

const StringBuffer ClientSettings::addressNodeName("Address");
const StringBuffer ClientSettings::displayNodeName("DisplayName");
const StringBuffer ClientSettings::nodePropertyName("value");
const StringBuffer ClientSettings::pendingSyncConfigPropertyName("pending-sync-config");
const StringBuffer ClientSettings::syncUsingPropertyName("syncUsing");
const StringBuffer ClientSettings::lastEmailSyncTimestampPropertyName("last-email");
const StringBuffer ClientSettings::lastEmailSyncStatusPropertyName("last-email-status");
const StringBuffer ClientSettings::schedulerMinutesPropertyName("schedulerMinutes");
const StringBuffer ClientSettings::cpIntervalPropertyName("cpInterval");
const StringBuffer ClientSettings::iapNamePropertyName("iapName");
const StringBuffer ClientSettings::firstRunPropertyName("firstRun");

/**
 * This private class is needed to get the phone IMEI. Anyone would think that grabbing
 * the IMEI is a matter of few lines of code. But this is Symbian! You MUST have
 * an acrive object to get it working. Noone knows why, but this is how it is.
 */
class CIMEI : public CActive
{
public:
    CIMEI() : CActive(EPriorityStandard) {}

    // Construction
    void CIMEI::ConstructL() {
        iPkg = new (ELeave) CTelephony::TPhoneIdV1Pckg(iV1);
        iTelephony = CTelephony::NewL();
        CActiveScheduler::Add(this);
    }

    // Destruction
    ~CIMEI() {
        Cancel(); // if any request outstanding, calls DoCancel() to cleanup
        delete iTelephony;
        delete iPkg;
    }

    static CIMEI* NewL() {
        CIMEI* self =new(ELeave) CIMEI();
        CleanupStack::PushL(self);
        self->ConstructL();
        CleanupStack::Pop(self);
        return self;
    }

    // Issue request: retrieve IMEI
    void GetIMEI(TRequestStatus& aStatus) {
        aStatus=KRequestPending;
        iIMEIStatus = &aStatus;
        iTelephony->GetPhoneId( iStatus, *iPkg );
        SetActive();
        CActiveScheduler::Start();
    }

    void CIMEI::RunL() {
        if ( (iStatus == KErrNone) ) {
            retrievedIMEI = (*iPkg)().iSerialNumber; // Read IMEI from package buffer
        }
        CActiveScheduler::Stop();

        //Signal complete and pass back result of CActive object
        User::RequestComplete(iIMEIStatus,iStatus.Int());
    }

    void CIMEI::DoCancel() {
        Cancel();
    }

private:
    CTelephony* iTelephony; // telephony object we own
    CTelephony::TPhoneIdV1 iV1;
    CTelephony::TPhoneIdV1Pckg* iPkg;
    TRequestStatus *iIMEIStatus;

public:
    TBuf<50> retrievedIMEI;
};



ClientSettings* ClientSettings::getInstance() {
    if (instance == NULL) {
        instance = new ClientSettings();
    }
    return instance;
}

ClientSettings::ClientSettings() : DMTClientConfig(),
                                   emailAddress     (""),
                                   displayName      (""),
                                   pendingSyncConfig(false),
                                   syncUsing        (ESyncUsingPush),
                                   lastEmailSyncTimestamp(0),
                                   lastEmailSyncStatus(ELastSyncNotFound),
                                   schedulerMinutes (30),
                                   cpInterval       (300),
                                   iapName          (""),
                                   firstRun         (false),
                                   dirty            (false),
                                   nativeDirty      (false),
                                   copiedAccountFromProfile(false)
{
}



StringBuffer ClientSettings::getImeiL() {

    // Get the phone IMEI
    CIMEI* imei = CIMEI::NewL();
    CleanupStack::PushL(imei);

    TRequestStatus status;
    imei->GetIMEI(status);

    // wait for request to complete
    User::WaitForRequest(status);

    // Check if IMEI retrieved correctly
    StringBuffer imeiSb;
    if (status==KErrNone) {
        TBuf<50> imeiStr(imei->retrievedIMEI);
        imeiSb = bufToStringBuffer(imeiStr);
    }
    CleanupStack::PopAndDestroy();

    return imeiSb;
}

StringBuffer ClientSettings::retrieveDeviceId() 
{
    StringBuffer newDevId("");
    StringBuffer imei;
    TRAPD(err, imei = getImeiL(); );
    
    if (err == KErrNone) {
        // We got the IMEI
        newDevId = "IMEI:";
        newDevId += imei;
        LOG.debug("New device ID=%s", newDevId.c_str());
    } 
    else {
        LOG.error("Cannot grab phone IMEI, the device Id is left emtpy");
    }
    return newDevId;
}


//
// --------------- Set methods ---------------
//
void ClientSettings::setPendingSyncConfig(const bool aPendingSyncConfig) {
    if (aPendingSyncConfig != pendingSyncConfig) {
        dirty = true;
        pendingSyncConfig = aPendingSyncConfig;
    }
}

void ClientSettings::setSyncUsing(const TSyncUsing aSyncUsing) {
    if (aSyncUsing != syncUsing) {
        dirty     = true;
        syncUsing = aSyncUsing;
    }
}

void ClientSettings::setLastEmailSyncTimestamp(TInt64 aTime) {
    if (aTime != lastEmailSyncTimestamp) {
        dirty = true;
        lastEmailSyncTimestamp = aTime;
    }
}

void ClientSettings::setLastEmailSyncStatus(TLastSyncStatus aStatus) {
    if (aStatus != lastEmailSyncStatus) {
        lastEmailSyncStatus = aStatus;
        dirty               = true;
    }
}

void ClientSettings::setSchedulerMinutes(const TInt aSchedulerMinutes) {
    if (schedulerMinutes != aSchedulerMinutes) {
        schedulerMinutes = aSchedulerMinutes;
        dirty = true;
    }
}

void ClientSettings::setCpInterval(const TInt aCpInterval) { 
    if (cpInterval != aCpInterval) {
        cpInterval = aCpInterval;
        dirty      = true;
    }
}

void ClientSettings::setIapName(const StringBuffer& aIapName) {
    if (iapName != aIapName) {
        iapName = aIapName;
        dirty   = true;
        nativeDirty = true;
    }
}

void ClientSettings::setFirstRun(const bool aFirstRun) {
    if (firstRun != aFirstRun) {
        firstRun = aFirstRun;
        dirty    = true;
    }
}

void ClientSettings::setPassword(const StringBuffer& aPassword) {
    const StringBuffer& password = getAccessConfig().getPassword();
    if (password != aPassword) {
        getAccessConfig().setPassword(aPassword);
        dirty = true;
        nativeDirty = true;
    }
}

void ClientSettings::setSyncURL(const StringBuffer& aSyncURL) {
    const StringBuffer& syncURL = getAccessConfig().getSyncURL();
    if (syncURL != aSyncURL) {
        getAccessConfig().setSyncURL(aSyncURL);
        
        // Also sw-update will reflect the new syncUrl.
        UpdateManager updateManager(NULL);
        updateManager.setUpdateURL(aSyncURL);
        
        dirty = true;
        nativeDirty = true;
    }
}

void ClientSettings::setUsername(const StringBuffer& aUsername) {
    const StringBuffer& username = getAccessConfig().getUsername();
    if (username != aUsername) {
        getAccessConfig().setUsername(aUsername);
        dirty = true;
        nativeDirty = true;
    }
}



void ClientSettings::forceDirtyOnMainNode() {
    dirty = true;
    nativeDirty = true;
}


bool ClientSettings::read() 
{
    bool ok = DMTClientConfig::read();
    if (!ok) {
        LOG.debug("DMTClientConfig::read failed");
        return false;
    }
    
    if (strlen(getSwv()) == 0) {
        LOG.debug("Swv empty -> config corrupted");
        return false;
    }
    
    
    LOG.debug("Reading Client settings");

    // Now load client specific settings
    StringBuffer context(APPLICATION_URI);

    //
    // Load the main node settings
    //
    DMTree mainTree(context.c_str());
    ManagementNode* mainNode = mainTree.getNode("");
    if (mainNode == NULL) {
        return false;
    }
    StringBuffer tmp(mainNode->readPropertyValue(pendingSyncConfigPropertyName.c_str()));
    pendingSyncConfig = (tmp == "true");
    
    tmp = mainNode->readPropertyValue(syncUsingPropertyName.c_str());
    if (tmp.length() > 0) {
        syncUsing = (TSyncUsing)atoi(tmp.c_str());
    }

    tmp = mainNode->readPropertyValue(lastEmailSyncTimestampPropertyName.c_str());
    // This piece of code is rather odd, but it looks like there are some
    // problems converting  between 32 and 64 bits value. For the timestamp we
    // don't really need to manipulate 64 bits values. We just need to store a
    // value at the end of the sync and to restore it. So we access the two
    // subwords by castint to 32bits.
    if (tmp.length() > 0) {
        TUint low = 0, high = 0;
        if (tmp.length() > 8) {
            StringBuffer sb  = tmp.substr(0, 8);
            StringBuffer sb1;
            sb1.sprintf("0x%s", sb.c_str());
            low = strtol(sb1.c_str(), NULL, 16);
            sb=tmp.substr(8, 8);
            sb1.sprintf("0x%s", sb.c_str());
            high = strtol(sb1.c_str(), NULL, 16);
        } else {
            // It fits in 8 bytes
            low = atoi(tmp.c_str());
        }
        TUint* p = (TUint*)&lastEmailSyncTimestamp;
        *p     = low;
        *(p+1) = high;
    }

    tmp = mainNode->readPropertyValue(lastEmailSyncStatusPropertyName.c_str());
    if (tmp.length() > 0) {
        lastEmailSyncStatus = (TLastSyncStatus)atoi(tmp.c_str());
    }

    tmp = mainNode->readPropertyValue(schedulerMinutesPropertyName.c_str());
    if (tmp.length() > 0) {
        schedulerMinutes = atoi(tmp.c_str());
    }

    tmp = mainNode->readPropertyValue(cpIntervalPropertyName.c_str());
    if (tmp.length() > 0) {
        cpInterval = atoi(tmp.c_str());
    }

    tmp = mainNode->readPropertyValue(iapNamePropertyName.c_str());
    if (tmp.length() > 0) {
        iapName = tmp;
    }
    delete mainNode;

    ok = readConfigSyncSourceSettings();
    return ok;
}

bool ClientSettings::isCredentialsEmpty()
{
    StringBuffer username(getUsername());
    StringBuffer password(getPassword());
    StringBuffer syncURL(getSyncURL());
    
    if (username.empty() || password.empty() || syncURL.empty()) {
        LOG.debug("Credentials empty");
        return true;
    }
    return false;
}


bool ClientSettings::readConfigSyncSourceSettings() {
    //
    // Load the ConfigSyncSource settings (meaning: handled by the config sync source)
    //
    StringBuffer context(APPLICATION_URI);
    context += configContext;
    DMTree emailTree(context.c_str());
    ManagementNode* addressNode = emailTree.getNode(addressNodeName.c_str());
    if (addressNode == NULL) {
        return false;
    }
    emailAddress = addressNode->readPropertyValue(nodePropertyName.c_str());
    delete addressNode;

    ManagementNode* displayNode = emailTree.getNode(displayNodeName.c_str());
    if (displayNode == NULL) {
        return false;
    }
    displayName = displayNode->readPropertyValue(nodePropertyName.c_str());
    delete displayNode;
    return true;
}



bool ClientSettings::save() {

    bool dmtSaved = DMTClientConfig::save();

    // If this config is not dirty, there is nothing to save
    if (!dirty) {
        return dmtSaved;
    }

    // We must propagate the information to the
    // native profiles
    StringBuffer username(getUsername());
    StringBuffer password(getPassword());
    StringBuffer syncURL(getSyncURL());
    TInt err;

    CSyncProfileManager* profileManager = NULL;
    CImapAccountManager* imapManager = NULL;
    if (nativeDirty) 
    {
        profileManager = CSyncProfileManager::createNewInstance();
        if (! profileManager->setCredentials(username, password, syncURL)) {
            LOG.error("Cannot set credentials in sync profile account");
        }

        if (ENABLE_SOURCE_MAIL) {
            TRAP(err,
                imapManager = CImapAccountManager::NewL();
                imapManager->setCredentials(username, password);
            );
            if (err != KErrNone) {
                LOG.error("Cannot create imap account manager");
            }
            // Propagate Display Name and Email Address
            imapManager->setAddress(emailAddress);
            imapManager->setDisplayName(displayName);
        }
    }

    // Now save client specific settings
    StringBuffer context(APPLICATION_URI);

    //
    // Save the main node settings
    //
    DMTree mainTree(context.c_str());
    ManagementNode* mainNode = mainTree.getNode("");
    if (mainNode == NULL) {
        return false;
    }
    mainNode->setPropertyValue(pendingSyncConfigPropertyName.c_str(),
                               pendingSyncConfig ? "true" : "false");
    StringBuffer tmp;
    tmp.sprintf("%d", syncUsing);
    mainNode->setPropertyValue(syncUsingPropertyName.c_str(), tmp.c_str());

    // In Symbian 64 bits values are not printed properly by our APIs. To
    // workaround this problem, we print the subwords (the endianess is not
    // important here as we just store/load values, do not manipulate them).
    TUint* p = (TUint*)&lastEmailSyncTimestamp;
    tmp.sprintf("%08x%08x", *p, *(p+1));
    mainNode->setPropertyValue(lastEmailSyncTimestampPropertyName.c_str(),
                               tmp.c_str());

    tmp.sprintf("%d", lastEmailSyncStatus);
    mainNode->setPropertyValue(lastEmailSyncStatusPropertyName.c_str(),
                               tmp.c_str());

    tmp.sprintf("%d", schedulerMinutes);
    mainNode->setPropertyValue(schedulerMinutesPropertyName.c_str(), tmp.c_str());
    
    tmp.sprintf("%d", cpInterval);
    mainNode->setPropertyValue(cpIntervalPropertyName.c_str(), tmp.c_str());
   
    mainNode->setPropertyValue(iapNamePropertyName.c_str(), iapName.c_str());
    // The IAP information must be propagated to the native profiles
    if (nativeDirty) {
        profileManager->setIap(iapName);
        if (ENABLE_SOURCE_MAIL) {
            imapManager->setIap(iapName);
        }
        delete profileManager;
        delete imapManager;
        nativeDirty = false;
    }
    delete mainNode;

    //
    // Save the ConfigSyncSource settings (meaning: handled by the config sync source)
    //
    context += configContext;
    DMTree emailTree(context.c_str());
    ManagementNode* addressNode = emailTree.getNode(addressNodeName.c_str());
    if (!emailAddress.empty()) {
        addressNode->setPropertyValue(nodePropertyName.c_str(), emailAddress.c_str());
    }
    delete addressNode;

    ManagementNode* displayNode = emailTree.getNode(displayNodeName.c_str());
    if (!displayName.empty()) {
        displayNode->setPropertyValue(nodePropertyName.c_str(), displayName.c_str());
    }
    delete displayNode;
    dirty = false;

    return false;
}


void ClientSettings::createDefaultConfig()
{
    //
    // AccessConfig
    //
    AccessConfig* ac = DefaultSymbianConfigFactory::getAccessConfig();
    setAccessConfig(*ac);
    delete ac;

    //
    // DeviceConfig
    //
    DeviceConfig* dc = DefaultSymbianConfigFactory::getDeviceConfig();
    setDeviceConfig(*dc);
    delete dc;
    
    //
    // SyncSources config
    //
    RArray<TInt> sourceIDs;
    sourceIDs.Reset();
    sourceIDs.Append(SOURCE_CALENDAR);
    sourceIDs.Append(SOURCE_CONFIG);
    sourceIDs.Append(SOURCE_CONTACTS);
    if (ENABLE_SOURCE_MAIL) {
        sourceIDs.Append(SOURCE_MAIL);
    }
    if (ENABLE_SOURCE_NOTES) {
        sourceIDs.Append(SOURCE_NOTES);
    }
    
    // Let's remove previously created sourceConfigs (if any) and reinitialize it.
    // It may happen we have empty/corrupted config files for sources, so remove them.
    if (sourceConfigs) { 
        delete [] sourceConfigs; 
        sourceConfigs = NULL;
    }
    sourceConfigsCount = 0;
    
    for (int i=0; i<sourceIDs.Count(); i++) {
        StringBuffer sourceName = getSourceNameFromID(sourceIDs[i]);
        SyncSourceConfig* sc = DefaultSymbianConfigFactory::getSyncSourceConfig(sourceName.c_str());
        addSyncSourceConfig(*sc);
        delete sc;
    }
    
    
    //
    // Root params
    //
    setFirstRun              (true);        // if no config found, it's the first time
    setPendingSyncConfig     (true);
    setSyncUsing             ((TSyncUsing)DEFAULT_SYNC_USING);
    setLastEmailSyncTimestamp(0);
    setLastEmailSyncStatus   (ELastSyncNotFound);
    setSchedulerMinutes      (DEFAULT_SCHEDULER_INTERVAL);
    setCpInterval            (CP_INTERVAL_SECONDS);
    setIapName               ("");

    //
    // Config syncsource params
    //
    emailAddress = "";
    displayName  = "";
    
    //
    // Set deviceID + software version + user agent.
    //
    StringBuffer devId = retrieveDeviceId();
    getDeviceConfig().setDevID(devId.c_str());
    
    StringBuffer swv = getPropertyFromUpgradeDir(PROPERTY_SOFTWARE_VERSION);
    getDeviceConfig().setSwv(swv.c_str());

    StringBuffer userAgent = createUserAgent(swv);
    getAccessConfig().setUserAgent(userAgent.c_str());
    
    
    // *** since 7.1.12 ***
    // Check if SYNC_PROFILE_NAME profile exists. If so, get account data from there.
    // This avoids slow syncs in case the user has already a SYNC_PROFILE_NAME profile in sync.
    // If profile not found we force a slow-sync for PIM, to avoid weird issues first time.
    // Note: we must do it before saving the config (it creates the profile)
    CSyncProfileManager* profileManager = CSyncProfileManager::createNewInstance();
    bool profileExists = profileManager->profileExists();
    StringBuffer username, password, syncURL;

    if (profileExists) {
        // Set the account data stored in the existing profile.
        setUsername(profileManager->getUsername());
        setPassword(profileManager->getPassword());
        setSyncURL (profileManager->getServerURI());
        copiedAccountFromProfile = true;
        
        // Check if all tasks (sources) are configured!
        // It may happen that 'notes' is never used nor configured, so will be created here.
        profileManager->CheckAllTasks();
    }
    else {
        // Reset timestamps for PIM sources.
        SyncSourceConfig* ssc = getSyncSourceConfig(SOURCE_CONTACTS_NAME);
        if (ssc) { ssc->setLast(0); }
        ssc = getSyncSourceConfig(SOURCE_CALENDAR_NAME);
        if (ssc) { ssc->setLast(0); }
        ssc = getSyncSourceConfig(SOURCE_NOTES_NAME);
        if (ssc) { ssc->setLast(0); }
        copiedAccountFromProfile = false;
    }
    delete profileManager;
    
    
    // To make sure we spread the settings to sync profile / email account.
    forceDirtyOnMainNode();

    sourceIDs.Close();
}


StringBuffer ClientSettings::getPropertyFromUpgradeDir(const StringBuffer& propertyName)
{
    StringBuffer value("");
    if (propertyName.empty()) {
        return value;
    }
    
    StringBuffer context(APPLICATION_URI);
    context += upgradeContext;
    DMTree upgradeTree(context.c_str());
    ManagementNode* upgradeNode = upgradeTree.getNode("");
    if (!upgradeNode) {
        return value;
    }
    value = upgradeNode->readPropertyValue(propertyName.c_str());
    
    delete upgradeNode;
    return value;
}


bool ClientSettings::makeUpgradeActions() 
{
    StringBuffer oldSwv = getDeviceConfig().getSwv();
    StringBuffer newSwv = getPropertyFromUpgradeDir(PROPERTY_SOFTWARE_VERSION);
    
    if (newSwv == oldSwv) {
        // No upgrade necessary.
        return false;
    }
    
    int oldVersion = getBuildNumberFromVersion(oldSwv.c_str());
    int newVersion = getBuildNumberFromVersion(newSwv.c_str());
    
    if (newVersion == oldVersion) {
        // No upgrade necessary.
        return false;
    }
    else if (newVersion < oldVersion) {
        // Downgrade software! -> avoid? warn user?
        LOG.info("Warning: the version installed is older than the previous one: '%s'", oldSwv.c_str());
    }
    
    //
    // Upgrade configuration.
    // **********************
    //
    getDeviceConfig().setSwv(newSwv.c_str());
    
    StringBuffer userAgent = createUserAgent(newSwv);
    getAccessConfig().setUserAgent(userAgent.c_str());
    
    //
    // Add here any code to ugrade the client based on the previous version.
    //
    if (oldVersion < 70104) 
    {
        // Remove the obsolete log file: new log name is defined in customization.h
        LOG.debug("Removing old log file");
        RFs fsSession;
        fsSession.Connect();
        RBuf name;
        name.Assign(charToNewBuf("C:\\data\\synclog.txt"));
        fsSession.Delete(name);
        fsSession.Close();
        name.Close();
    }

    
    LOG.info("Configuration succesfully upgraded to new version: '%s'", newSwv.c_str());
    return true;
}


StringBuffer ClientSettings::createUserAgent(const StringBuffer& swv)
{
    StringBuffer userAgent("");
    userAgent.sprintf("%s v. %s", APPLICATION_NAME, swv.c_str());
    return userAgent;
}

bool ClientSettings::createSyncMLProfile()
{
    StringBuffer username  = getUsername();
    StringBuffer password  = getPassword();
    StringBuffer serverURI = getSyncURL();

    CSyncProfileManager* profileManager = CSyncProfileManager::createNewInstance();
    bool ret = profileManager->createProfile(username, password, serverURI, iapName);
    
    delete profileManager;
    return ret;
}

bool ClientSettings::checkCopiedAccountFromProfile()
{
    return copiedAccountFromProfile;
}

bool ClientSettings::createIMAPAccount()
{
    StringBuffer username  = getUsername();
    StringBuffer password  = getPassword();
    
    CImapAccountManager* imapManager = NULL;
    TRAPD(err, imapManager = CImapAccountManager::NewL());
    if (err != KErrNone) {
        LOG.error("Cannot open IMAP account manager (err =  %d)", err);
        return false;
    }
    
    /*if (imapManager->AccountExists(SYNC_PROFILE_NAME)) {
        // No need to create the IMAP account
        delete imapManager;
        return false;
    }*/
    
    // This call creates the IMAP account
    bool ret = imapManager->setCredentials(username, password);
    
    // Propagate all parameters
    imapManager->setAddress(emailAddress);
    imapManager->setDisplayName(displayName);
    imapManager->setIap(iapName);
    
    delete imapManager;
    return ret;
}


/* --- Not used: dirs are created automatically by DMNode if fopen fails ---
TInt ClientSettings::createAllDirs(const StringBuffer& rootDir, RArray<TInt>& sourceIDs)
{
    TInt err = KErrNone;
    
    // Connect to the file server
    RFs fileSession;
    err = fileSession.Connect();
    if (err) {
        LOG.error("Cannot connect to the file server (code %d)", err);
        return err;
    }
    
    CleanupClosePushL(fileSession);
    
    StringBuffer mainPath(rootDir);
    mainPath += APPLICATION_URI;
    
    StringBuffer syncmlPath(mainPath);
    syncmlPath += CONTEXT_SPDS_SYNCML;
    
    StringBuffer sourcesPath(mainPath);
    sourcesPath += CONTEXT_SPDS_SOURCES;
    
    //
    // "\spds\syncml" dirs
    //
    err = makeDir(syncmlPath, CONTEXT_AUTH,       &fileSession);
    if (err) { goto finally; }
    err = makeDir(syncmlPath, CONTEXT_CONN,       &fileSession);
    if (err) { goto finally; }
    err = makeDir(syncmlPath, CONTEXT_DEV_DETAIL, &fileSession);
    if (err) { goto finally; }
    err = makeDir(syncmlPath, CONTEXT_DEV_INFO,   &fileSession);
    if (err) { goto finally; }
    err = makeDir(syncmlPath, CONTEXT_EXT,        &fileSession);
    if (err) { goto finally; }
    
    //
    // "\spds\sources" dirs
    //
    for (int i=0; i<sourceIDs.Count(); i++) {
        StringBuffer sourceName = getSourceNameFromID(sourceIDs[i]);
        err = makeDir(sourcesPath, sourceName.c_str(), &fileSession);
        if (err) { goto finally; }
    }

finally:
    CleanupStack::PopAndDestroy(&fileSession);
    return err;
}

TInt ClientSettings::makeDir(const StringBuffer& aPath, const char* aContext, RFs* aFileSession)
{
    TInt err = KErrNone;
    RBuf pathDes;
    
    StringBuffer path(aPath);
    concatDirs(path, aContext);
    path = contextToPath(path.c_str());     // "/" -> "\"
    
    path += "\\";
    LOG.debug("mkdir: '%s'", path.c_str());
    pathDes.Assign(stringBufferToNewBuf(path));
    err = aFileSession->MkDirAll(pathDes);
    pathDes.Close();
    
    return err;
}
*/


