/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2008 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

#include "base/globalsdef.h"
#include "base/Log.h"
#include "spdm/ManagementNode.h"
#include "spdm/DMTree.h"
#include "client/CacheSyncSource.h"
#include "spds/AbstractSyncSourceConfig.h"
#include "base/util/ArrayListEnumeration.h"

#include "ConfigSyncSource.h"

USE_NAMESPACE

ConfigSyncSource::ConfigSyncSource(const WCHAR* name,
                                   const StringBuffer& applicationUri,
                                   AbstractSyncSourceConfig* sc,
                                   KeyValueStore* cache) :
                  CacheSyncSource(name, sc, cache),
                  applicationUri(applicationUri) {
}

ConfigSyncSource::~ConfigSyncSource() {  
}

void ConfigSyncSource::setMimeType(StringBuffer& t) {
    mimeType = t;
}

const StringBuffer& ConfigSyncSource::getMimeType() const {
    return mimeType;
}

void* ConfigSyncSource::getItemContent(StringBuffer& key, size_t* size) {

    StringBuffer* keysIter = (StringBuffer*)keysList.front();
    StringBuffer* itemsIter = (StringBuffer*)itemsList.front();
    while(keysIter) {
        if (*keysIter == key) {
            *size = strlen(itemsIter->c_str());
            return stringdup(itemsIter->c_str());
        }
        keysIter  = (StringBuffer*)keysList.next();
        itemsIter = (StringBuffer*)itemsList.next();
    }
    *size = 0;
    return NULL;
}

char* ConfigSyncSource::readItemContent(WCHAR* key, size_t* size) {
    // The key is the node name
    StringBuffer keyString;
    keyString.convert(key);
    if (keyString.find("./") == 0) {
        keyString = keyString.substr(2);
    }
    DMTree tree(applicationUri.c_str());
    ManagementNode* node = tree.getNode(keyString.c_str());
    if (getLastErrorCode() == ERR_DM_TREE_NOT_AVAILABLE) {
        // First time the node is not found, it's generated empty.
        resetError();
    }
    if (!node) {
        // There is something really wrong here...
        return NULL;
    }
    char* value = node->readPropertyValue("value");
    LOG.debug("ConfigSyncSource: %s=%s", keyString.c_str(), value);
    delete node;
    *size = strlen(value);
    return value;
}

int ConfigSyncSource::removeAllItems() {
    cleanItem(PUSH_STATUS_FIELD);
    cleanItem(EMAIL_FILTER_STATUS_FIELD);
    cleanItem(EMAIL_ADDRESS_FIELD);
    cleanItem(VISIBLE_NAME_FIELD);

    return 0;
}

void ConfigSyncSource::cleanItem(WCHAR* key) {

    StringBuffer keyString;
    keyString.convert(key);
    if (keyString.find("./") == 0) {
        keyString = keyString.substr(2);
    }
    DMTree tree(applicationUri.c_str());
    ManagementNode* node = tree.getNode(keyString.c_str());
    if (!node) {
        // There is something really wrong here...
        return;
    }
    node->setPropertyValue("value", "");
    delete node;
}
    
Enumeration* ConfigSyncSource::getAllItemList() {

    itemsList.clear();
    keysList.clear();
    size_t size;

    // Read all items and discard the empty ones. We should use a map here
    StringBuffer pushStatus(readItemContent(PUSH_STATUS_FIELD, &size));
    StringBuffer emailFilterStatus(readItemContent(EMAIL_FILTER_STATUS_FIELD, &size));
    StringBuffer emailAddress(readItemContent(EMAIL_ADDRESS_FIELD, &size));
    StringBuffer visibleName(readItemContent(VISIBLE_NAME_FIELD, &size));

    if (!pushStatus.empty()) {
        StringBuffer key;
        key.convert(PUSH_STATUS_FIELD);
        keysList.add(key);
        itemsList.add(pushStatus);
    }
    if (!emailFilterStatus.empty()) {
        StringBuffer key;
        key.convert(EMAIL_FILTER_STATUS_FIELD);
        keysList.add(key);
        itemsList.add(emailFilterStatus);
    }
    if (!emailAddress.empty()) {
        StringBuffer key;
        key.convert(EMAIL_ADDRESS_FIELD);
        keysList.add(key);
        itemsList.add(emailAddress);
    }
    if (!visibleName.empty()) {
        StringBuffer key;
        key.convert(VISIBLE_NAME_FIELD);
        keysList.add(key);
        itemsList.add(visibleName);
    }

    // Collect all local items (their keys)
    ArrayListEnumeration* enumeration = new ArrayListEnumeration(keysList);
    return enumeration;
}
    
int ConfigSyncSource::insertItem(SyncItem& item) {
    // The key is the node name
    DMTree tree(applicationUri.c_str());
    StringBuffer key;
    key.convert(item.getKey());
    if (key.find("./") == 0) {
        key = key.substr(2);
    }
    ManagementNode* node = tree.getNode(key.c_str());
    if (getLastErrorCode() == ERR_DM_TREE_NOT_AVAILABLE) {
        // First time the node is not found, it's generated empty.
        resetError();
    }
    if (!node) {
        return STC_COMMAND_FAILED;
    }
    LOG.debug("ConfigSyncSource::insertItem: %s", (char*)item.getData());
    node->setPropertyValue("value", (char*)item.getData());
    delete node;
    return STC_OK;
}

int ConfigSyncSource::modifyItem(SyncItem& item) {
    // Same behavior as for insertItem
    return insertItem(item);
}

int ConfigSyncSource::removeItem(SyncItem& item) {
    // This should never happen
    return STC_COMMAND_FAILED;
}


