/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2008 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */


#include <eikenv.h>
#include <aknglobalnote.h>
#include "base/util/symbianUtils.h"
#include "base/util/timeUtils.h"
#include "base/SymbianLog.h"
#include "base/util/PropertyFile.h"
#include "base/adapter/PlatformAdapter.h"
#include "spds/SyncManager.h"
#include "client/SyncClient.h"
#include "updater/Updater.h"

#include "FunambolAppView.h"
#include "commonDefs.h"
#include "customization.h"
#include "ClientSettings.h"
#include "SyncScheduler.h"
#include "ClientPush.h"
#include "UpdateManager.h"
#include "SyncActiveObject.h"
#include "utils.h"
#include "FakeSyncEngine.h"

/// Number of applications available to sync: contacts, calendar, notes.
#define NUM_SYNC_APPS                   3


using namespace Funambol;

CSyncEngine* CSyncEngine::NewL(CFunambolAppUi* aAppUI)  { 
        CSyncEngine* self = CSyncEngine::NewLC(aAppUI);
        CleanupStack::Pop(self);

        return self;
    }
 
CSyncEngine* CSyncEngine::NewLC(CFunambolAppUi* aAppUI) { 
        CSyncEngine* self = new (ELeave) CSyncEngine;
        CleanupStack::PushL(self);
        self->ConstructL(aAppUI);

        return self;
    } 


void CSyncEngine::ConstructL(CFunambolAppUi* aAppUI)
    {
    iAppUI = aAppUI;
    
    iCurrentSource = 0;
    iQueuedJobs.Reset();

    // We must originate and complete an asynchronous requests
    // inside the same thread, so we save its ID.
    RThread thisThread;
    iMainThreadID = thisThread.Id();
    LOG.debug("thread id = %d", iMainThreadID);
    
    // Adds this object to the Active Scheduler
    CActiveScheduler::Add(this);
    
    // Creates the scheduler.
    iScheduler = CSyncScheduler::NewL(this);
    }


void CSyncEngine::syncConfig()
{
    // We don't want more than 1 sync instance
    if (SyncActiveObject::CheckSyncInProgress()) {
        LOG.info("Config sync already in progress");
        return;
    }
    
    // Initialize the Cache File Name
    StringBuffer fullPath(PlatformAdapter::getHomeFolder());
    StringBuffer cacheFile;
    cacheFile.sprintf("%s%s", fullPath.c_str(), "config_sync_source.dat");

    TRAPD(err, SyncActiveObject::syncConfigLD(this, cacheFile));
}

void CSyncEngine::StartSync(const RArray<TInt> &aSourceIDList, const TSyncRunners from)
{
    //
    // Starts the sync of all sources, as separate jobs.
    // Each job has only 1 task to synchronize.
    iNumSources = aSourceIDList.Count();
    if (aSourceIDList.Count() > 0) {
        // This is important to set the first source as the current one,
        // during the initial "connecting" phase (before the job really starts)
        SetCurrentSource((TInt)aSourceIDList[0]);
        for (TInt i=0; i<aSourceIDList.Count(); i++) {
            StartSync((TInt)aSourceIDList[i]);
        }
    }
}

void CSyncEngine::StartSync(const TInt aSourceID)
{
    if (aSourceID == SOURCE_CONFIG) {
        syncConfig();
    }
}

void CSyncEngine::RunL()
{
    LOG.info("RunL - status = %d, state = %d", iStatus.Int(), iAsyncAction);
    if(iStatus == KErrNone) 
    {
        switch(iAsyncAction)
        {
            case EStartSync:
            {
                StartSync(iRequestedSourceList, ESyncTriggeredAsyncAction);
                iRequestedSourceList.Reset();
                break;
            }
            case EStopSync:
            {
                StopSync();
                break;
            }
            case EEnableScheduler:
            {
                // Currently not used.
                break;
            }
            default:
                break;
        }
        iAsyncAction = EIdle;
    }
    else {
        LOG.error("CSyncEngine::RunL() error: status = %d", iStatus.Int());
    }
}

RArray<TInt> CSyncEngine::GetActiveSourcesList() { 
        RArray<TInt> list;
        list.AppendL(0);

        return list;
}

TInt CSyncEngine::RunError(TInt aError) {
    Funambol::LOG.error("CSyncEngine RunL() error: %d", aError);
    return 0;
}

void CSyncEngine::syncConfigEnded(TInt status)
{
    // Update the App UI, reporting the error if necessary
    iAppUI->UpdateSourceStatus(SOURCE_CONFIG, ESyncStatusCompleted, status);
    iAppUI->ConfigSyncEnded();
}   


