/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2008 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

#include "base/SymbianLog.h"
#include "base/symbianadapter.h"
#include "base/util/symbianUtils.h"
#include "base/util/stringUtils.h"
#include "base/FConnection.h"
#include "push/CTPService.h"

#include "utils.h"
#include "commonDefs.h"
#include "customization.h"
#include "PushEngine.h"
#include "ClientSettings.h"

using namespace Funambol;


CPushEngine* CPushEngine::NewL(CFunambolAppUi* aAppUI)
{
    CPushEngine* self = CPushEngine::NewLC(aAppUI);
    CleanupStack::Pop(self);
    return self;
}

CPushEngine* CPushEngine::NewLC(CFunambolAppUi* aAppUI)
{
    CPushEngine* self = new(ELeave) CPushEngine();
    self->ConstructL(aAppUI);
    CleanupStack::PushL(self);
    return self;
}


CPushEngine::CPushEngine() 
{
    iAppUI      = NULL;
    iClientPush = NULL;
}


void CPushEngine::ConstructL(CFunambolAppUi* aAppUI)
{
    iAppUI = aAppUI;
    CTPService* ctp = CTPService::getInstance();
    
    if (ctp->getConfig()->getCtpPort() == 0) {
        // Config corrupted (or missing) -> generate default and save it.
        createDefaultCTPConfig();
        ctp->getConfig()->saveCTPConfig();
    }
    
    // Now receive notifications for push messages.
    ctp->registerPushListener(*this);
    
    // Create the ClientPush. Don't start the listeners now.
    TRAPD(err, iClientPush = CClientPush::NewL(iAppUI);)
    if (err) {
        LOG.error("Error initializing ClientPush: %d", err);
        iAppUI->DisplayMessage(err, EErrorLevel);
    }
}

CPushEngine::~CPushEngine() 
{
    CTPService* ctp = CTPService::getInstance();
    ctp->stopCTP();
    
    delete iClientPush;
}


void CPushEngine::start()
{
    LOG.info("Starting CTP service...");
    CTPService* ctp = CTPService::getInstance();
    ctp->startCTP();
    
    // Starts listening for CP notifications.
    iClientPush->start();
}

void CPushEngine::stop() 
{
    LOG.info("Stopping CTP service...");
    CTPService* ctp = CTPService::getInstance();
    ctp->stopCTP();
    
    // Stops listening for CP notifications.
    iClientPush->stop();
}

void CPushEngine::restart() 
{
    stop();
    start();
}


TInt CPushEngine::getPushState()
{
    CTPService* ctp = CTPService::getInstance();
    return ctp->getCtpState();
}

bool CPushEngine::isPushRunning()
{
    TInt state = getPushState();
    
    if ( (state > CTPService::CTP_STATE_DISCONNECTED) && (iClientPush->isRunning()) ) {
        // Both up and running.
        return true;
    }
    return false;
}


void CPushEngine::onNotificationReceived(const ArrayList& serverURIList) 
{
    LOG.info("Push notification received: %d sources", serverURIList.size());
    RArray<TInt> sourceList;
    
    // Convert each ServerURI string into a source ID.
    for (int i=0; i < serverURIList.size(); i++) {
        StringBuffer* uri = (StringBuffer*)serverURIList[i];
        if ( !uri || !(uri->length()) ) {
            continue;
        }
        StringBuffer sourceName = getSourceNameFromServerURI(*uri);
        TInt sourceID = getSourceIDFromName(sourceName);
        if (sourceID != -1) {
            sourceList.Append(sourceID);
        }
    }
    
    if (sourceList.Count() > 0) {
        // Start the push-triggered sync!
        iAppUI->StartSync(sourceList, ESyncTriggeredPush);
    }
    sourceList.Close();
}


void CPushEngine::onCTPError(const int errorCode, const int additionalInfo)
{
    LOG.debug("PushEngine: CTP error occurred: code %d", errorCode);
    
    switch (errorCode) 
    {
        case CTPService::CTP_ERROR_NOT_AUTHENTICATED:
        case CTPService::CTP_ERROR_UNAUTHORIZED:
        case CTPService::CTP_ERROR_AUTH_FORBIDDEN:
        {
            // Should not happen: auth is checked during config sync...
            iAppUI->DisplayMessage(ERR_AUTH_FAILED, EErrorLevel);
            break;
        }
        case CTPService::CTP_ERROR_CONNECTION_FAILED:
        {
            // It happens only when the max CTPRetry time is reached (15 min)
            iAppUI->DisplayMessage(ERR_CTP_NO_CONNECTION, EErrorLevel);
            break;
        }
        default:
            // Ignore other errors.
            break;
    }
}


StringBuffer CPushEngine::getSourceNameFromServerURI(const StringBuffer& uri)
{
    StringBuffer sourceName, nodeName;
    
    //
    // Access the config nodes for sources
    //
    nodeName.sprintf("%s/%s", APPLICATION_URI, CONTEXT_SPDS_SOURCES);
    DMTree* dmt = DMTreeFactory::getDMTree(APPLICATION_URI);
    ManagementNode *node = dmt->readManagementNode(nodeName.c_str());
    if (!node) {
        setErrorF(ERR_INVALID_CONTEXT, ERRMSG_INVALID_CONTEXT, nodeName.c_str());
        LOG.error("getSourceIDFromServerURI: error: %s", getLastErrorMsg());
        goto finally;
    }
    
    //
    // Find the source name from its serverURI.
    //
    for(int i=0; i<node->getChildrenMaxCount(); i++) {
        ManagementNode *src = node->getChild(i);
        char *tmp = src->readPropertyValue(PROPERTY_SOURCE_URI);
        //LOG.debug("i:%d uri:%s tmp:%s", i, uri, tmp);
        if (tmp && !strcmp(uri.c_str(), tmp)) {
            char* name = src->readPropertyValue(PROPERTY_SOURCE_NAME);
            sourceName = name;
            delete [] name;
            delete [] tmp;
            break;
        }
        delete [] tmp;
    }
    

    /*
    * If the client has modified the source name, in the push the server continue
    * to send i.e. scard for the contact, scal for the calendar and stask for the task.
    * mail for the mail. So if the previous loop doesn't find anything, try in this way
    */
    if (sourceName.length() == 0) 
    {
        if (uri == "mail") {
            sourceName = SOURCE_MAIL_NAME;
        }
        else if ((uri == "scard")  || (uri == "card") || (uri == "oxcard")) {
            sourceName = SOURCE_CONTACTS_NAME;
        }
        else if ( (uri == "scal")  || (uri == "cal")  || (uri == "event") ||
                  (uri == "stask") || (uri == "task") || 
                  (uri == "oxcal") || (uri == "oxtask") ) {
            sourceName = SOURCE_CALENDAR_NAME;
        } 
        else if ((uri == "snote")  || (uri == "note")) {
            sourceName = SOURCE_NOTES_NAME;
        }
    }

finally:
    delete dmt;
    delete node;
    return sourceName;
}


void CPushEngine::createDefaultCTPConfig()
{
    LOG.debug("Generating default CTP configuration...");
    
    CTPConfig* ctpConfig = CTPService::getInstance()->getConfig();
    if (!ctpConfig) {
        return;
    }

    ClientSettings* config = ClientSettings::getInstance();
    StringBuffer hostName = ctpConfig->getHostName(config->getSyncURL());
    
    ctpConfig->setUrlTo         (hostName);
    ctpConfig->setCtpCmdTimeout (60);
    ctpConfig->setCtpConnTimeout(0);
    ctpConfig->setCtpPort       (CTP_PORT);
    ctpConfig->setCtpReady      (CTP_READY_INTERVAL);
    ctpConfig->setCtpRetry      (5);
    ctpConfig->setMaxCtpRetry   (900);
    ctpConfig->setNotifyTimeout (120);
    
    // -- not used --
    //ctpConfig->setQueuePush   (true);
    //ctpConfig->setPush        (1);
    //ctpConfig->setPolling     (0);
}



