/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2008 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

#include <eikenv.h>

#include "base/util/StringBuffer.h"
#include "base/util/symbianUtils.h"
#include "base/util/timeUtils.h"
#include "base/Log.h"
#include "base/util/PropertyFile.h"
#include "spds/SyncManager.h"
#include "client/SyncClient.h"

#include "commonDefs.h"
#include "customization.h"
#include "ConfigSyncSource.h"

#include "ClientSettings.h"
#include "utils.h"
#include "SyncActiveObject.h"

SyncActiveObject* SyncActiveObject::syncAO = NULL;
TimeoutTimer* SyncActiveObject::tt = NULL;
CSyncEngine*  SyncActiveObject::iSyncEngine = NULL;
SyncActiveObject::State SyncActiveObject::iState = ENone;


class TimeoutTimer : public CActive
{
public:

    static TimeoutTimer* TimeoutTimer::NewL()
    {
        TimeoutTimer* pSelf = new(ELeave) TimeoutTimer();
        CleanupStack::PushL(pSelf);
        pSelf->ConstructL();
        CleanupStack::Pop();

        return pSelf;
    }

    void TimeoutTimer::Start()
    {
        Cancel();

        iTimer.After(iStatus, CREDENTIALS_CHECK_TIMEOUT * 1000000);
        SetActive();
    }

    TimeoutTimer::~TimeoutTimer()
    {
        Cancel();
        iTimer.Close();
    }

private:

    TimeoutTimer::TimeoutTimer() : CActive(EPriorityStandard),
                                   iCancelled(EFalse)
    {
    }

    void TimeoutTimer::ConstructL()
    {
        iTimer.CreateLocal();
        CActiveScheduler::Add(this);
    }

    void TimeoutTimer::RunL()
    {
        if (!iCancelled) {
            LOG.debug("TimeoutTimer triggered, ending config sync");
            SyncActiveObject::syncConfigEnded(true, 500);
        }
        // Self destroy
        delete this;
    }

    void TimeoutTimer::DoCancel()
    {
        iTimer.Cancel();
        iCancelled = ETrue;
    }

private:
    RTimer            iTimer;
    TBool             iCancelled;
};



SyncActiveObject* SyncActiveObject::NewL(StringBuffer cacheFile)
{
    SyncActiveObject* self = SyncActiveObject::NewLC(cacheFile);
    CleanupStack::Pop(self);
    return self;
}

SyncActiveObject* SyncActiveObject::NewLC(StringBuffer cacheFile)
{
    SyncActiveObject* self = new(ELeave) SyncActiveObject(cacheFile);
    self->ConstructL();
    CleanupStack::PushL(self);
    return self;
}


void SyncActiveObject::syncConfigLD(CSyncEngine* aSyncEngine, StringBuffer cacheFile)
{
    iSyncEngine = aSyncEngine;
    
    // We don't want more than 1 sync instance
    if (CheckSyncInProgress()) {
        LOG.info("Config sync already in progress");
        return;
    }

    LOG.info("Starting config sync (timeout = %d seconds)", CREDENTIALS_CHECK_TIMEOUT);
    syncAO = SyncActiveObject::NewL(cacheFile);
    syncAO->sync();

    LOG.debug("Starting timeout timer");
    tt = TimeoutTimer::NewL();
    tt->Start();
}

void SyncActiveObject::syncConfigEnded(bool timeout, TInt status)
{
    LOG.debug("Sync Config Ended");
    if (!timeout) {
        LOG.debug("Sync config terminated normally");
        tt->Cancel();
        syncAO = NULL;
    } else {
        // Timeout
        LOG.debug("Sync config terminated on timeout");
        if (syncAO) {
            syncAO->Stop();
        }
    }

    // If we get a timeout, but the sync has already succesfully finished, then
    // we simply ignore it
    if (!(timeout && syncAO == NULL)) {
        iSyncEngine->syncConfigEnded(status);
    }
}


bool SyncActiveObject::CheckSyncInProgress()
{
    if (iState == ESync) {
        return true;
    }
    return false;
}


SyncActiveObject::SyncActiveObject(StringBuffer cacheFile)
                                       : CActive(CActive::EPriorityStandard),
                                         iCacheFile(cacheFile),
                                         status(500)
{
}

void SyncActiveObject::ConstructL()
{
    CActiveScheduler::Add(this);
    iState = ENone;
}

SyncActiveObject::~SyncActiveObject()
{
    Cancel();
}

void SyncActiveObject::Stop() {
    iState = EStopped;
    Cancel();
}

void SyncActiveObject::sync()
{
    if (iState != EStopped) {
        LOG.debug("SyncActiveObject about to start");
        TRequestStatus* status = &iStatus;
        User::RequestComplete(status, KErrNone);
        iState = ESync;
        SetActive();
    } else {
        LOG.debug("The sync cannot start as it has been stopped already");
    }
}

void SyncActiveObject::syncL(void)
{
    ClientSettings* config = ClientSettings::getInstance();

    PropertyFile* cache = new PropertyFile(iCacheFile.c_str());

    //
    // Create the SyncSource passing its name and its config.
    //
    wchar_t* wConfigSourceName = toWideChar(SOURCE_CONFIG_NAME);
    StringBuffer ssPath;
    ssPath.sprintf("%s/", APPLICATION_URI);
    ConfigSyncSource source(wConfigSourceName, ssPath,
                            config->getSyncSourceConfig(SOURCE_CONFIG_NAME), cache);
    SyncSource* ssArray[] = { &source, NULL };
    delete [] wConfigSourceName;

    // We remove all the old values, as they are going to be replaced. This is
    // already done by doing a refresh from server, but in order to be sure
    // we do it also here
    source.removeAllItems();
    source.setPreferredSyncMode(SYNC_REFRESH_FROM_SERVER);

    //
    // Create the SyncClient .
    //
    SyncClient configClient;
    
    //
    // Make sure we have an active connection - if not, retry 3 times asking the user.
    //
    if (!startConnection(MAX_CONNECTION_RETRY_ASKING_USER)) {
        return;
    }

    //
    // Start the sync!
    //
    status = configClient.sync(*config, ssArray);

    if (status) {
        LOG.error("Error in config sync");
    } 
    else {
        // Save config params only if sync successful!

        // This is a bit tricky. The sync changed the configuration on disk,
        // but the config object has not been updated. We need to reread the
        // config in order to refresh the config (which acts like a cache)
        config->readConfigSyncSourceSettings();
        // This is needed because we need to push the parameters into the
        // native configuration (the ConfigSyncSource updates only the DM)
        config->save();
    }

    // Print sync results.
    StringBuffer res("");
    configClient.getSyncReport()->toString(res);
    LOG.info("\n%s", res.c_str());
}

void SyncActiveObject::DoCancel(void)
{
    LOG.debug("SyncActiveObject::DoCancel");
    iState = ECancelled;
}


void SyncActiveObject::RunL(void)
{
    LOG.debug("SyncActiveObject::RunL");

    switch(iState) {
        case ESync:
        {
            // Start the sync
            syncL();

            // Move to the done status (if we were not
            // cancelled)
            if (iState == ESync) {
                LOG.debug("iState is still ESync");
                iState = EDone;
            }
            TRequestStatus* s = &iStatus;
            User::RequestComplete(s, KErrNone);
            SetActive();
            break;
        }
        case ECancelled:
            LOG.debug("SyncActiveObject Cancelled");
            // Delete this
            delete this;
            break;
        case EDone:
            LOG.debug("SyncActiveObject Done");
            syncConfigEnded(false, status);
            // Delete this
            delete this;
            break;
        default:
            break;
    }
}

