/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2008 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

#include <eikenv.h>
#include <aknglobalnote.h>
#include <syncmlerr.h>

#ifdef FAKE_IMAP_ACCOUNT
#include "FakeImapEmailEngine.h"
#else
#include "ImapEmailEngine.h"
#endif

#include "base/util/symbianUtils.h"
#include "base/util/timeUtils.h"
#include "base/SymbianLog.h"
#include "base/util/PropertyFile.h"
#include "base/adapter/PlatformAdapter.h"
#include "spds/SyncManager.h"
#include "client/SyncClient.h"
#include "updater/Updater.h"

#include "FunambolAppView.h"
#include "commonDefs.h"
#include "customization.h"
#include "ClientSettings.h"
#include "SyncScheduler.h"
#include "ClientPush.h"
#include "UpdateManager.h"
#include "SyncActiveObject.h"
#include "utils.h"

#include "SyncEngine.h"

/// Number of applications available to sync: contacts, calendar, notes.
#define NUM_SYNC_APPS                   3



using namespace Funambol;

typedef struct SyncApp {
    TInt          sourceId;       /**< The corresponding source ID */
    TBuf<32>      dataSource;     /**< The dataSource for this task */
} SyncApps;

static SyncApps SyncTasks[NUM_SYNC_APPS] = {
//  sourceId                dataSource
//  -------------------------------------------
    {SOURCE_CONTACTS,    _L("C:Contacts.cdb")},
    {SOURCE_CALENDAR,    _L("C:Calendar")    },
    {SOURCE_NOTES,       _L("c:Notepad.dat") }
};



CSyncEngine* CSyncEngine::NewL(CFunambolAppUi* aAppUI)
{
    CSyncEngine* self = CSyncEngine::NewLC(aAppUI);
    CleanupStack::Pop(self);
    return self;
}

CSyncEngine* CSyncEngine::NewLC(CFunambolAppUi* aAppUI)
{
    CSyncEngine* self = new(ELeave) CSyncEngine();
    CleanupStack::PushL(self);
    self->ConstructL(aAppUI);
    return self;
}

/*
 * Class constructor
 */
CSyncEngine::CSyncEngine() : CActive(CActive::EPriorityStandard), 
                             iActiveProfileID(-1), 
                             iCreateProfileRetry(0)
{
}


void CSyncEngine::ConstructL(CFunambolAppUi* aAppUI)
{
    iAppUI = aAppUI;
    
    iCurrentSource = 0;
    iQueuedJobs.Reset();

    // We must originate and complete an asynchronous requests
    // inside the same thread, so we save its ID.
    RThread thisThread;
    iMainThreadID = thisThread.Id();
    //Funambol::LOG.debug("thread id = %d", iMainThreadID);
    
    // Adds this object to the Active Scheduler
    CActiveScheduler::Add(this);
    
    // Opens SyncML session.
    InitializeSessionL();
    
    // Creates the scheduler.
    iScheduler = CSyncScheduler::NewL(this);

    // Email engine is built on demand
    emailEngine = NULL;
}

void CSyncEngine::InitializeSessionL()
{
    //
    // open syncML session
    //
    Funambol::LOG.debug("Opening SyncML session...");
    iSyncMLSession.OpenL();   


  /*// *Note*: THIS IS MANDATORY TO USE THE SAME SESSION FROM DIFFERENT THREADS!
    // After calling this function the RSyncMLSession object 
    // may be used by threads other than than the one that created it.
    TInt err = iSyncMLSession.ShareAuto();
    if (err != KErrNone) {
        Funambol::LOG.error("CSyncEngine error: unable to share RSyncMLSession (code %d)", err);
        User::Leave(ERR_SHARE_SYNCML_SESSION);
    }*/

    // Request event notifications
    iSyncMLSession.RequestProgressL(*this);
    iSyncMLSession.RequestEventL(*this);

    iActiveProfileID = -1;
}

TInt CSyncEngine::getActiveProfileID() {

    if (iActiveProfileID != -1) {
        return iActiveProfileID;
    }

    // get the list of available profiles
    RArray<TSmlProfileId> profilesArray;
    iSyncMLSession.ListProfilesL(profilesArray, ESmlDataSync);
    
    RBuf profileName;
    Funambol::LOG.debug("Loading Sync profile...");
    profileName.Assign(wcharToNewBuf(WSYNC_PROFILE_NAME));

    // search for desired profile
    for(TInt count=0; count<profilesArray.Count(); ++count)
    {
        RSyncMLDataSyncProfile profile;
        profile.OpenL(iSyncMLSession, profilesArray[count], ESmlOpenReadWrite);
        if (profile.DisplayName().Compare(profileName) == 0)
        {
            iActiveProfileID = profile.Identifier();
            profile.Close();
            LOG.debug("Profile found: id = %d", iActiveProfileID);
            break;
        }
        profile.Close();
        
        /* Feature not supported... :(
        RSyncMLSettings settings;
        settings.OpenL(iSyncMLSession);
        settings.SetValueL(ESmlSettingDisplayBgResult, ESmlDisplayAllResults);
        settings.Close(); */
    }
    profileName.Close();
    profilesArray.Close();

 
    if(iActiveProfileID == -1) {
        //
        // No active profile found
        //
        if (iCreateProfileRetry < 1) {
            // Create the profile and try again (recursive call, only 1 time!)
            LOG.info("SyncML profile '%s' not found: creating it...", SYNC_PROFILE_NAME);
            iCreateProfileRetry ++;
            if (ClientSettings::getInstance()->createSyncMLProfile()) {
                getActiveProfileID();
            }
            else {
                LOG.error("Cannot create SyncML profile"); 
            }
        }
        else {
            LOG.error("SyncML profile '%s' not found", SYNC_PROFILE_NAME);
        }
    }
    return iActiveProfileID;
}


/*
 * Class destructor
 */
CSyncEngine::~CSyncEngine()
{
    //Funambol::LOG.debug("Closing CSyncEngine");
    CActive::Cancel();
    iSyncMLSession.Close();
    
    iQueuedJobs.Close();
    iRequestedSourceList.Close();
    
    delete iScheduler;
}


/*
 * Sync an array of sources (exclude config)
 */
void CSyncEngine::StartSync(const RArray<TInt> &aSourceIDList, const TSyncRunners from)
{
    LOG.debug("CSyncEngine::StartSync - from %d, %d sources", from, aSourceIDList.Count());
    
    //
    // Here we take different actions, based on who triggered the startsync.
    //
    switch(from)
    {
        case ESyncTriggeredManual:
        {
            // ****** MANUAL SYNC ******
            if (CheckSyncInProgress()) {
                // A sync is already running
                iAppUI->DisplayMessage(INFO_SYNC_IN_PROGRESS, EInfoLevel);
                return;
            }
            Funambol::LOG.debug("StartSync: manual (%d sources)", aSourceIDList.Count());
            break;
        }
        case ESyncTriggeredPush:
        {
            // ****** CTP PUSH ******
            // We MUST use the active-object, because the sync action has
            // to be executed inside the same thread that opened the SyncML session.
            // Pass the list of sourceID through the internal parameter 'iRequestedSourceList'.
            mergeRArray(iRequestedSourceList, aSourceIDList);
            RequestAsyncAction(EStartSync);
            Funambol::LOG.debug("StartSync: CTP push (%d sources)", aSourceIDList.Count());
            return;
        }
        case ESyncTriggeredClientPush:
        {
            // ****** CLIENT PUSH ******
            // We discard the CP sync if the source is under sync and
            // during 'receiving' status...
            if (isSourceUnderSyncAndReceiving((TInt)aSourceIDList[0])) {
                LOG.debug("CP notification discarded: source %d is under sync", aSourceIDList[0]);
                return;
            }
            ClientSettings* config = ClientSettings::getInstance();
            TInt interval = config->getCpInterval();
            
            // Set the source list for the scheduler, merge the list.
            iScheduler->setSourceList(aSourceIDList, true);
            
            // Enable ripetitive scheduler for CP: if sync fails, it will be
            // rescheduled again. If sync success, scheduler will be disabled 
            // (so outstanding syncs will be discarded).
            iScheduler->enable(true, interval);
            LOG.debug("StartSync: CP push (%d sources)", aSourceIDList.Count());
            return;
        }
        case ESyncTriggeredAsyncAction:
        case ESyncTriggeredScheduled:
        {
            // ****** From Active Objects RunL() ******
            Funambol::LOG.debug("StartSync: from active object (%d sources)", aSourceIDList.Count());
            break;
        }
        default:
            return;
    }
    
    //
    // Starts the sync of all sources, as separate jobs.
    // Each job has only 1 task to synchronize.
    iNumSources = aSourceIDList.Count();
    if (aSourceIDList.Count() > 0) {
        // This is important to set the first source as the current one,
        // during the initial "connecting" phase (before the job really starts)
        SetCurrentSource((TInt)aSourceIDList[0]);
        for (TInt i=0; i<aSourceIDList.Count(); i++) {
            StartSync((TInt)aSourceIDList[i]);
        }
    }
}



// Syncs a single data source. Private method.
void CSyncEngine::StartSync(const TInt aSourceID)
{
    LOG.debug("CSyncEngine::StartSync (source id is %d)", aSourceID);
    
    // Safe check: the source syncmode MUST NOT be "none".
    // (it may happen, for CP or CTP push)
    if (getSourceSyncMode(aSourceID) == ESyncModeNone) {
        LOG.debug("Aborting sync of '%s' (source is disabled).", getSourceNameFromID(aSourceID).c_str());
        return;
    }
    
    if (aSourceID == SOURCE_MAIL) {
        syncEmail();
        return;
    }

    if (aSourceID == SOURCE_CONFIG) {
        syncConfig();
        return;
    }
    
    // Check if the current source is already in the queued jobs.
    // In that case it's not necessary to start queue another job, 
    // so we discard this sync.
    if (findSourceIDInQueuedJobs(aSourceID)) {
        LOG.debug("source %d is already queued, discard.", aSourceID);
        return;
    }
    
    // Read the syncType from config.
    ClientSettings* config = ClientSettings::getInstance();
    StringBuffer sourceName = getSourceNameFromID(aSourceID);
    SyncSourceConfig* ssc = config->getSyncSourceConfig(sourceName.c_str());
    if (!ssc) {
        LOG.error("StartSync aborted: config not found for source '%s'", sourceName.c_str());
        return;
    }
    
    TSmlSyncType syncType;
    if (ssc->getLast() == 0) {
        if (aSourceID == SOURCE_NOTES) { 
            // SlowSync is not supported for Notes!! :-0
            LOG.info("Starting a TWO-WAY SYNC for '%s' (can't force a slow sync)", sourceName.c_str());
            syncType = ESmlTwoWay;
        }
        else {
            // Last timestamp is 0: this means it's the first sync, or the source was cleaned up.
            LOG.info("Starting a SLOW SYNC for '%s'", sourceName.c_str());
            syncType = ESmlSlowSync;
        }
    }
    else {
        syncType = getSyncTypeFromText(ssc->getSync());
    }
    


    // Check if the SyncML profile exists.
    // This must be done before retrieving the task ID.
    if (getActiveProfileID() == KErrNotFound) {
        iAppUI->DisplayMessage(ERR_PROFILE_NOT_FOUND, EErrorLevel);
        return;
    }

    // Get the corresponding task from the list of tasks of active profile
    TSmlTaskId taskID = GetTaskIDFromSourceID(aSourceID);
    LOG.debug("task id = %d", taskID);

    if(taskID == KErrNotFound) {
        LOG.error("Internal error: sync task not found for source '%s'", sourceName.c_str());
        return;
    }
    else {
        // Applications are tasks; every profile has a list of tasks associated
        // We use a specific array of tasks "required", with only 1 taskId.
        RArray<TSmlTaskId> rTaskArray;
        rTaskArray.Append(taskID);
        
        // Start the new job of synchronization
        RSyncMLDataSyncJob job;
        TRAPD(err, job.CreateL(iSyncMLSession, getActiveProfileID(), syncType, rTaskArray))
        rTaskArray.Close();
        if (err) {
            LOG.error("Error creating RSyncMLDataSyncJob: code %d", err);
            return;
        }
        
        // Save the <jobID, sourceID> entry in the FIFO list of queued jobs
        AddQueuedJob(job.Identifier(), aSourceID);
        
        job.Close();
    }
}


TInt CSyncEngine::StopSync()
{
    LOG.debug("CSyncEngine::StopSync");
    
    // We can have only one sync at a time running
    // So we first check what sync is in progress
    if (emailEngine && emailEngine->CheckSyncInProgress()) {
        emailEngine->CancelOperation();
        // Don't return, as there could be another source in progress
        // TODO: remove this comment when Email source will be blocking!
        //return 1;
    }

    if (getActiveProfileID() == -1) {
        // No sync profile, available, there is nothing we can stop
        return 0;
    }

    TSmlJobId jobID = GetCurrentJobID();
    if (jobID == -1) {
        return 0;    // Nothing to do
    }
    
    // Stop the running job.
    TInt error = StopJob(jobID);
    
    // Stop all the queued jobs (mark them to abort before start).
    StopQueuedJobs();
    
    if (!error) {
        return 1;   // Ok, job stopped.
    }
    return error;
}


TBool CSyncEngine::CheckSyncInProgress() 
{
    // Check for a config sync in progress.
    // During config sync the UI is locked and all push services
    // are disabled. Therefore a sync of the config should never
    // overlap to any other sync
    if (SyncActiveObject::CheckSyncInProgress()) {
        //LOG.debug("Config sync is in progress");
        return ETrue;
    }
    
    if (getActiveProfileID() != -1) {
        TSmlJobId jobID = GetCurrentJobID();
        if (jobID != -1) {
            return ETrue;   // sync is running
        }
    }
    // Check if we have email sync in progress
    if (emailEngine && emailEngine->CheckSyncInProgress()) {
        return ETrue;
    }
    return EFalse;
}

TSmlJobId CSyncEngine::GetCurrentJobID() 
{
    TSmlJobId jobID;
    TSmlUsageType jobType;
    
    // Returns '-1' if no job running!  Note: Doc is wrong! :(
    TRAPD(err, iSyncMLSession.CurrentJobL(jobID, jobType);)
    if (err) {
        Funambol::LOG.error("Error getting job running: %d", err);
        return KErrNotFound;
    }

    return jobID;
}



void CSyncEngine::OnSyncMLSyncProgress(TStatus aStatus, TInt aInfo1, TInt aInfo2)
{
    // Save the actual syncml status, used in OnSyncMLDataSyncModifications()
    switch(aStatus)
    {
        // These events are forwarded to AppUI, for proper actions
        // Here is a mapping between 'TStatus' defined by Symbian SyncML and our 'TCurrentSyncStatus'.
        case ESmlConnecting:
            iCurrentSyncStatus = ESyncStatusConnecting;
            LOG.info("Connecting to SyncML Server...");
            break;
        case ESmlConnected:
            iCurrentSyncStatus = ESyncStatusConnected;
            LOG.debug("Connected to SyncML Server");
            break;
        case ESmlDisconnected:
            iCurrentSyncStatus = ESyncStatusDisconnected;
            LOG.debug("Disconnected from SyncML Server");
            break;
        case ESmlLoggingOn:
            iCurrentSyncStatus = ESyncStatusLoggingOn;
            LOG.debug("Logging on...");
            break;
        case ESmlLoggedOn:
            iCurrentSyncStatus = ESyncStatusLoggedOn;
            LOG.debug("Logged on");
            break;
        case ESmlSendingModificationsToServer:
            iCurrentSyncStatus = ESyncStatusSendingData;
            LOG.info("Sending modifications...");
            break;
        case ESmlReceivingModificationsFromServer:
            iCurrentSyncStatus = ESyncStatusReceivingData; 
            LOG.info("Receiving modifications...");
            break;
        case ESmlSendingMappingsToServer:
            iCurrentSyncStatus = ESyncStatusMappings;
            LOG.info("Sending mappings...");
            break;        
        case ESmlCompleted:
            LOG.debug("Synchronization ended.");
            // NOTE: don't update the UI, we still don't know if there was an error.
            //       A SyncMLSessionEvent will be generated for this.
            return;

        // Other events are ignored!
        default:
            return;
    }
    
    // Send the notification to the AppUI, to resfresh the View.
    iAppUI->UpdateSourceStatus(iCurrentSource, iCurrentSyncStatus);
}


void CSyncEngine::OnSyncMLDataSyncModifications(TInt aTaskId,
                                                const TSyncMLDataSyncModifications& aClientMod,
                                                const TSyncMLDataSyncModifications& aServerMod)
{

    if (iCurrentSyncStatus == ESyncStatusSendingData) {
        //
        // Client -> Server
        //
        iTotalServerItems += aServerMod.iNumAdded + aServerMod.iNumDeleted + 
                             aServerMod.iNumMoved + aServerMod.iNumFailed  + aServerMod.iNumReplaced;
        
        iAppUI->UpdateSourceStatus(iCurrentSource, ESyncStatusSendingItemNumber, iTotalServerItems);
    }
    else if (iCurrentSyncStatus == ESyncStatusReceivingData) {
        //
        // Server -> Client
        //
        iTotalClientItems += aClientMod.iNumAdded + aClientMod.iNumDeleted + 
                             aClientMod.iNumMoved + aClientMod.iNumFailed  + aClientMod.iNumReplaced;
        
        iAppUI->UpdateSourceStatus(iCurrentSource, ESyncStatusReceivingItemNumber, iTotalClientItems);
    }

}

void CSyncEngine::OnSyncMLSyncError(TErrorLevel aErrorLevel, TInt aError, TInt aTaskId, TInt aInfo1, TInt aInfo2)
{
    LOG.error("SyncML sync error occurred: code %d", aError);
    iCurrentSyncStatus = ESyncStatusError;
    iAppUI->UpdateSourceStatus(iCurrentSource, ESyncStatusError, aError);

    // Not here: works only on N91.
    // This call is moved to OnSyncSessionEvent()
    //ManageSyncMLErrors(aError, aInfo1);
}


void CSyncEngine::ManageSyncMLErrors(const TInt aError, const TInt aAdditionalData)
{
    switch(aError)
    {
        case SyncMLError::KErrTransportRequest:
        case SyncMLError::KErrTransportHeader:
        case SyncMLError::KErrTransportSvrNoUri:
        case SyncMLError::KErrTransportSvrError:
        case SyncMLError::KErrTransportDisconnected:
        case SyncMLError::KErrConnectionError:
        case SyncMLError::KErrCannotCommunicateWithServer:
        case SyncMLError::KErrPhoneOnOfflineMode:
        {
            LOG.error("Connection error (code %d). Stopping the sync.", aError);
            StopQueuedJobs();
            iAppUI->DisplayMessage(ERR_CONNECTING, EErrorLevel);
            break;
        }
        case SyncMLError::KErrTransportAuthenticationFailure:
        case SyncMLError::KErrAuthenticationFailure:
        {
            LOG.error("Authentication failed (code %d). Stopping the sync.", aError);
            StopQueuedJobs();
            iAppUI->DisplayMessage(ERR_AUTH_FAILED, EErrorLevel);
            break;
        }
        case SyncMLError::KErrServerDatabaseNotFound:
        {
            LOG.error("Source not found on Server (code %d).", aError);
            iAppUI->DisplayMessage(ERR_SOURCE_NOT_FOUND, EErrorLevel);
            break;
        }
        case SyncMLError::KErrDatabaseAdapterIndexError:
        case SyncMLError::KErrSessionIdNotValid:
        case SyncMLError::KErrVerDTDMissing:
        case SyncMLError::KErrProtoMissing:
        case SyncMLError::KErrTargetLocMissing:
        case SyncMLError::KErrMappingInfoMissing:
        case SyncMLError::KErrMappingInfoNotAcked:
        case SyncMLError::KErrVersionDTDNotSupported:
        case SyncMLError::KErrProtoNotSupported:
        case SyncMLError::KErrStatusCommandForCommandNotSent:
        case SyncMLError::KErrStatusCommandNotTheSame:
        case SyncMLError::KErrMissingStatusCommands:
        case SyncMLError::KErrFailedToAddRecord:
        case SyncMLError::KErrFailedToReplaceRecord:
        case SyncMLError::KErrFailedToDeleteRecord:
        case SyncMLError::KErrObjectTooLargeToSendToServer:
        case SyncMLError::KErrUserInteractionFailed:
        case SyncMLError::KErrStatusSizeMismatch:
        case SyncMLError::KErrFailedToCloseDatabase:
        case SyncMLError::KErrGeneral:
        case SyncMLError::KErrProfileAccessDenied:
        case SyncMLError::KErrObjectInUse:
        case SyncMLError::KErrServerIdNotMatching:
        {
            LOG.error("SyncML error (code %d).", aError);
            iAppUI->DisplayMessage(ERR_SYNCML_GENERIC_FAILURE, EErrorLevel);
            break;
        }
        case SyncMLError::KErrSourceLocMissing:
        case SyncMLError::KErrClientDatabaseNotFound:
        {
            LOG.error("Source not found on Client (code %d).", aError);
            iAppUI->DisplayMessage(ERR_SYNCML_GENERIC_FAILURE, EErrorLevel);
            break;
        }
        case SyncMLError::KErrSlowSync:
        {
            // TODO: display a warning popup?
            LOG.info("Server requested a SLOW SYNC for '%s'", getSourceNameFromID(iCurrentSource).c_str());
            break;
        }
        default:
        {
            LOG.error("Unexpected SyncML error (code %d)", aError);
            iAppUI->DisplayMessage(ERR_SYNCML_GENERIC_FAILURE, EErrorLevel);
            break;
        }
    }
}


void CSyncEngine::OnSyncMLSessionEvent(TEvent aEvent, TInt aIdentifier, TInt aError, TInt aAdditionalData)
{
    LOG.debug("SyncML session event: event=%d, id=%d, err=%d, data=%d", 
                                     aEvent, aIdentifier, aError, aAdditionalData);
    
    switch (aEvent)
    {
        case EJobStart:
        {
            if (aError) {
                LOG.debug("SyncML session error occurred on JobStart: code %d", aError);
                iCurrentSyncStatus = ESyncStatusError;
                iAppUI->UpdateSourceStatus(iCurrentSource, ESyncStatusError, aError);
                // Should we display a msg for these errors?
                //ManageSyncMLErrors(aError, aAdditionalData);
            }
            else {
                // 'aIdentifier' is new job ID.
                // Here we manage the starting of a new source synced (new job).
                ManageNewJobStarted(aIdentifier);
            }
            return;
        }
        
        case EJobStop:
        {
            if (aError) {
                if (aError == KErrCancel) {
                    // User canceled the sync
                    LOG.info("Sync cancelled by the user (%d items sent, %d items received)", 
                             iTotalServerItems, iTotalClientItems);
                    iCurrentSyncStatus = ESyncStatusCanceled;
                }
                else {
                    // Generic failure
                    LOG.info("Sync ended with error (%d items sent, %d items received)", 
                             iTotalServerItems, iTotalClientItems);
                    iCurrentSyncStatus = ESyncStatusNotCompleted;
                    ManageSyncMLErrors(aError, aAdditionalData);
                }
            }
            else {
                // OK. The sync has ended succesfully.
                iCurrentSyncStatus = ESyncStatusCompleted;
                LOG.info("Sync ended succesfully! (%d items sent, %d items received)", 
                         iTotalServerItems, iTotalClientItems);
            }
            iAppUI->UpdateSourceStatus(iCurrentSource, iCurrentSyncStatus, aError);
            --iNumSources;
            checkForUpdates();
            return;
        }
        
        default:
        {
            // Other events: just log errors
            if (aError) {
                LOG.error("SyncML session error: event = %d, error = %d", aEvent, aError);
            }
            return;
        }
    
    }
}



void CSyncEngine::ManageNewJobStarted(const TSmlJobId aJobID)
{
    TInt index = GetQueuedJobIndex(aJobID);
    if (index != KErrNotFound) 
    {
        // It's just started a new job: we save the current source ID
        // to know which task is actually under sync.
        SetCurrentSource(iQueuedJobs[index].sourceID);
        
        if (iQueuedJobs[index].toAbort) {
            // This job need to be interrupted! Stop it now.
            LOG.debug("The job %d needs to be aborted", aJobID);
            StopJob(aJobID);
        }
        else {
            Funambol::LOG.info("**** Starting sync job of '%s' ****", getSourceNameFromID(iCurrentSource).c_str());
            // Reset counters
            iTotalServerItems = 0;
            iTotalClientItems = 0;
        }

        // IMPORTANT! Job just started, remove it from the queued list.
        // This also avoids the queued list growing...
        iCriticalQueuedJobs.Wait();
        iQueuedJobs.Remove(index);
        iCriticalQueuedJobs.Signal();
    }
}

TInt CSyncEngine::StopJob(const TSmlJobId aJobID)
{
    TInt err = KErrNone;
    
    if (aJobID >= 0) {
        RSyncMLDataSyncJob job;
        TRAP(err, job.OpenL(iSyncMLSession, aJobID);)
        if (err) {
            Funambol::LOG.error("StopJob: error opening job %d, code %d", aJobID, err);
            return -1;
        }
        
        Funambol::LOG.debug("Stopping job (id = %d)", aJobID);
        TRAP(err, job.StopL();)
        if (err) {
            Funambol::LOG.error("StopJob: error stopping job, code %d", err);
        }
        job.Close();
    }
    return err;
}

void CSyncEngine::StopQueuedJobs()
{
    LOG.debug("Marking queued jobs to be aborted");
    
    iCriticalQueuedJobs.Wait();
    for (TInt i=0; i<iQueuedJobs.Count(); i++) {
        iQueuedJobs[i].toAbort = ETrue;
    }
    iCriticalQueuedJobs.Signal();

    // Delete also any queued job  ???? works? ????
//    RArray<TSmlJobId> queuedJobIDs;
//    TSmlUsageType jobType;
//    iSyncMLSession.ListQueuedJobsL(queuedJobIDs, jobType);
//    for (TInt i=0; i<queuedJobIDs.Count(); i++) {
//        if (queuedJobIDs[i] != KSmlNullId) {
//            RSyncMLDataSyncJob queuedJob;
//            queuedJob.OpenL(iSyncMLSession, queuedJobIDs[i]);
//            Funambol::LOG.debug("Removing job (id = %d) from queue", queuedJobIDs[i]);
//            queuedJob.StopL();    // StopL() also removes the job from the queue
//            queuedJob.Close();
//        }
//    }
}


TLastSyncStatus CSyncEngine::GetLastSyncStatus(const TInt aSourceID,
                                               TTime* aLastSyncTime,
                                               TInt* aLastSyncResult) 
{
    if (aSourceID == SOURCE_MAIL) {
        return GetLastEmailSyncStatus(aLastSyncTime, aLastSyncResult);
    } else {
        return GetLastPimSyncStatus(aSourceID, aLastSyncTime, aLastSyncResult);
    }
}


TLastSyncStatus CSyncEngine::GetLastEmailSyncStatus(TTime* aLastSyncTime,
                                                    TInt*  aLastSyncResult)
{
    TLastSyncStatus lastSyncStatus;
    ClientSettings* config = ClientSettings::getInstance();
    if (config->getLastEmailSyncTimestamp() == 0) 
    {
        // Never synchronized or unable to get the timestamp
        lastSyncStatus = ELastSyncNotFound;
    } 
    else 
    {
        TInt64 microsecs = config->getLastEmailSyncTimestamp();
        TTime tmp(microsecs);
        
        TTime nowUTC;
        nowUTC.UniversalTime();
        if (tmp > nowUTC) {
            // Fix: last sync time is in the future -> set to 'not found'
            return ELastSyncNotFound;
        }
        
        utcToLocalTime(tmp);         // it was a UTC value
        *aLastSyncTime = tmp;
        lastSyncStatus = (TLastSyncStatus)config->getLastEmailSyncStatus();
        (*aLastSyncResult) = config->getLastEmailSyncStatus();
    }
    return lastSyncStatus;
}



TLastSyncStatus CSyncEngine::GetLastPimSyncStatus(const TInt aSourceID,
                                                  TTime* aLastSyncTime,
                                                  TInt* aLastSyncResult) 
{
    //LOG.debug("CSyncEngine::GetLastPimSyncStatus for %s", getSourceNameFromID(aSourceID).c_str());
    TInt err = KErrNone;
    TLastSyncStatus lastSyncStatus = ELastSyncNotFound;

    TTime nowUTC;
    nowUTC.UniversalTime();

    // Get the corresponding task ID to search for
    TSmlTaskId taskID = GetTaskIDFromSourceID(aSourceID);
    //LOG.debug("task = %d", taskID);

    // Opens the historyLog on this profile.
    RSyncMLHistoryLog history;
    TRAP(err, history.OpenL(iSyncMLSession, getActiveProfileID());)
    if (err) {
        Funambol::LOG.error("Error opening RSyncMLHistoryLog: code %d", err);
        return ELastSyncNotFound;
    }

    // Sort the list by time, last entry is the most recent sync.
    history.SortEntries(CSyncMLHistoryEntry::ESortByTime);
    TInt numEntries = history.Count(); 

    TInt entryIndex = -1;
    RSyncMLDataSyncProfile profile;
    profile.OpenL(iSyncMLSession, getActiveProfileID(), ESmlOpenRead);
    if (err) {
        Funambol::LOG.error("Error opening RSyncMLDataSyncProfile: code %d", err);
        goto finally;
    }

    // Cycle through the entries, starting from the most recent one.
    // We want to find the latest sync job that contains this task.
    for (TInt i=numEntries-1; i>=0; i--) 
    {
        const CSyncMLHistoryJob* hjob = CSyncMLHistoryJob::DynamicCast(&(history.Entry(i)));
        TInt numTasks = hjob->TaskCount();
        
        // Check if this task was included in this sync job.
        for (int j=0; j<numTasks; j++) {
            if (hjob->TaskAt(j).iTaskId == taskID) {
                entryIndex = i;
                break;
            }
        }
        
        if (entryIndex >= 0) {
            //
            // Source was synced in the past
            //
            TTime lastSyncTimeUTC = hjob->FinishTime();     // returns an UTC value
            if (lastSyncTimeUTC > nowUTC) {
                // Fix: last sync time is in the future -> ignore this entry.
                entryIndex = -1;
                continue;
            }

            if (hjob->ResultCode() == 0) {
                //
                // Last sync successful: get last timestamp and set the out parameters
                //
                (*aLastSyncTime) = lastSyncTimeUTC;
                utcToLocalTime(*aLastSyncTime);
                
                lastSyncStatus = ELastSyncSuccessful;
                (*aLastSyncResult) = 0;
            }
            else {
                //
                // Last sync not completed
                //
                lastSyncStatus = ELastSyncFailed;
                (*aLastSyncResult) = hjob->ResultCode();
            }
            goto finally;
        }
    }

    //
    // If here, the source was never synced. Note: out parameters are not modified.
    //
    lastSyncStatus = ELastSyncNotFound;


finally:
    profile.Close();
    history.Close();
    return lastSyncStatus;
}


TSyncMode CSyncEngine::getSourceSyncMode(const TInt aSourceID)
{
    // Read the config
    StringBuffer name = getSourceNameFromID(aSourceID);
    ClientSettings* config = ClientSettings::getInstance();
    SyncSourceConfig* ssc = config->getSyncSourceConfig(name.c_str());
    if (!ssc) {
        LOG.error("Config not found for source '%s'", name.c_str());
        return ESyncModeNone;
    }
    
    StringBuffer syncMode = ssc->getSync();
    
    if (syncMode == SYNC_TWO_WAY_TEXT) {
        return ESyncModeNormal;
    }
    else if (syncMode == SYNC_ONE_WAY_FROM_CLIENT_TEXT) {
        return ESyncModeToServerOnly;
    }
    else if (syncMode == SYNC_ONE_WAY_FROM_SERVER_TEXT) {
        return ESyncModeToPhoneOnly;
    }
    else if (syncMode == SYNC_SLOW_TEXT) {
        return ESyncModeSlow;
    }
    else if (syncMode == SYNC_NONE_TEXT) {
        return ESyncModeNone;
    }
    
    return ESyncModeNormal;   // default
}

void CSyncEngine::setSourceSyncMode(const TInt aSourceID, const TSyncMode aSyncMode)
{
    StringBuffer name = getSourceNameFromID(aSourceID);
    ClientSettings* config = ClientSettings::getInstance();
    SyncSourceConfig* ssc = config->getSyncSourceConfig(name.c_str());
    if (!ssc) {
        LOG.error("Config not found for source '%s'", name.c_str());
        return;
    }
    
    switch (aSyncMode)
    {
        case ESyncModeNormal:
            ssc->setSync(SYNC_TWO_WAY_TEXT);
            break;
        case ESyncModeToServerOnly:
            ssc->setSync(SYNC_ONE_WAY_FROM_CLIENT_TEXT);
            break;
        case ESyncModeToPhoneOnly:
            ssc->setSync(SYNC_ONE_WAY_FROM_SERVER_TEXT);
            break;
        case ESyncModeSlow:
            ssc->setSync(SYNC_SLOW_TEXT);
            break;
        case ESyncModeNone:
            ssc->setSync(SYNC_NONE_TEXT);
            break;
        default:
            ssc->setSync(SYNC_TWO_WAY_TEXT);
            break;
    }
    
    config->save();
}

void CSyncEngine::setSourceTimestamp(const TInt aSourceID, const TUint timestamp)
{
    StringBuffer name = getSourceNameFromID(aSourceID);
    ClientSettings* config = ClientSettings::getInstance();
    SyncSourceConfig* ssc = config->getSyncSourceConfig(name.c_str());
    if (!ssc) {
        LOG.error("Config not found for source '%s'", name.c_str());
        return;
    }
    ssc->setLast(timestamp);
    config->save();
}



//void CSyncEngine::setSyncError(TErrorLevel aErrorLevel, TInt aError) {
//    
//    iLastErrorLevel = aErrorLevel;
//    iLastErrorCode = aError;
//    // Set the error message, given the code?
//}



//
//  -------  TODO: can we use a map?? -----------
//
TBuf<32> CSyncEngine::GetDataSourceFromSourceID(const TInt aSourceID) 
{
    for (int i=0; i<NUM_SYNC_APPS; i++) {
        if (SyncTasks[i].sourceId == aSourceID) {
            return (SyncTasks[i].dataSource);
        }
    }
    return SyncTasks[0].dataSource;
}

TInt CSyncEngine::GetSourceIDFromDataSource(const TDesC& aDataSource) 
{
    for (int i=0; i<NUM_SYNC_APPS; i++) {
        if (SyncTasks[i].dataSource == aDataSource) {
            return (SyncTasks[i].sourceId);
        }
    }
    return SyncTasks[0].sourceId;
}


TSmlTaskId CSyncEngine::GetTaskIDFromSourceID(const TInt aSourceID) 
{
    TInt err = KErrNone;
    TSmlTaskId taskID = KErrNotFound;
    
    // Open the active profile
    RSyncMLDataSyncProfile profile;
    TRAP(err, profile.OpenL(iSyncMLSession, getActiveProfileID(), ESmlOpenRead);)
    if (err) {
        Funambol::LOG.error("Error opening RSyncMLDataSyncProfile: code %d", err);
        return KErrNotFound;
    }
    
    // Get the active list of tasks
    RArray<TSmlTaskId> pTaskArray;
    TBuf<32> taskDataSource;
    TRAP(err, profile.ListTasksL(pTaskArray);)
    if (err) {
        LOG.error("Error getting task list: code %d", err);
        goto finally;
    }
    
    // Get the dataSource name of the task we want to synchronize
    taskDataSource = GetDataSourceFromSourceID(aSourceID);
    
    //
    // Search the desired task inside the profile task array
    //
    for (TInt count=0; count<pTaskArray.Count(); ++count) 
    {
        RSyncMLTask task;
        TRAP(err, task.OpenL(profile, pTaskArray[count]);)
        if (err) {
            Funambol::LOG.error("Error opening RSyncMLTask: code %d", err);
            goto finally;
        }
      
        // search for a particular task based on 'taskDataSource'
        if (task.ClientDataSource().Compare(taskDataSource) == 0) {
            taskID = task.Identifier();
            task.Close();
            break;
        }
        task.Close();
    }

finally:
    pTaskArray.Reset();
    profile.Close();
    return taskID;
}


TInt CSyncEngine::GetSourceIDFromTaskID(const TSmlTaskId aTaskID)
{
    TInt err = KErrNone;
    TInt sourceID = SOURCE_CONTACTS;
    
    // Open the active profile
    RSyncMLDataSyncProfile profile;
    TRAP(err, profile.OpenL(iSyncMLSession, getActiveProfileID(), ESmlOpenRead);)
    if (err) {
        Funambol::LOG.error("Error opening RSyncMLDataSyncProfile: code %d", err);
        return SOURCE_CONTACTS; // as a default
    }
    
    // Get the active list of tasks
    RArray<TSmlTaskId> pTaskArray;
    TRAP(err, profile.ListTasksL(pTaskArray);)
    if (err) {
        Funambol::LOG.error("Error getting task list: code %d", err);
        goto finally;
    }
    
    // Find the desired task inside the profile task array
    TInt index = pTaskArray.Find(aTaskID);
    if (index != KErrNotFound) 
    {
        // Open the task and get the dataSource name
        RSyncMLTask task;
        TRAP(err, task.OpenL(profile, pTaskArray[index]);)
        if (err) {
            Funambol::LOG.error("Error opening RSyncMLTask: code %d", err);
            goto finally;
        }
        
        sourceID = GetSourceIDFromDataSource(task.ClientDataSource());
        task.Close();
    }
    
finally:
    pTaskArray.Reset();
    profile.Close();
    return sourceID;
}



TInt CSyncEngine::AddQueuedJob(const TSmlJobId aJobID, const TInt aSourceID, const TBool aToAbort)
{
    JobEntry newJob;
    newJob.jobID    = aJobID;
    newJob.sourceID = aSourceID;
    newJob.toAbort  = aToAbort;
    
    iCriticalQueuedJobs.Wait();
    iQueuedJobs.Append(newJob);
    iCriticalQueuedJobs.Signal();

    return iQueuedJobs.Count();
}

TInt CSyncEngine::GetQueuedJobIndex(const TSmlJobId aJobID)
{
    for (TInt i=0; i<iQueuedJobs.Count(); i++) {
        if (iQueuedJobs[i].jobID == aJobID) {
            return i;
        }
    }
    return KErrNotFound;    // -1
}

TBool CSyncEngine::findSourceIDInQueuedJobs(const TInt aSourceID)
{
    for (TInt i=0; i<iQueuedJobs.Count(); i++) {
        if (iQueuedJobs[i].sourceID == aSourceID) {
            return ETrue;
        }
    }
    return EFalse;
}


TBool CSyncEngine::isSourceUnderSyncAndReceiving(const TInt aSourceID)
{
    // First: check on the Calendar source
    if (aSourceID == SOURCE_CALENDAR) 
    {
        // Calendar Change notification can arrive a bit later...
        // So we also check if a sync has just finished (error window = 2 seconds).
        TInt lastSyncResult = 0;
        TTime now, lasSyncTime;
        now.HomeTime();
        GetLastPimSyncStatus(aSourceID, &lasSyncTime, &lastSyncResult);
        if (lasSyncTime > (now - TTimeIntervalSeconds(2))) {
            LOG.debug("Calendar sync just finished");
            return ETrue;
        }
    }
    
    if (!CheckSyncInProgress()) {
        return EFalse;
    }
    
    if (aSourceID == iCurrentSource) {
        if ( iCurrentSyncStatus == ESyncStatusReceivingData       ||
             iCurrentSyncStatus == ESyncStatusReceivingItemNumber ||
             iCurrentSyncStatus == ESyncStatusMappings ) {      // Just to be sure, maybe a little delay...
            return ETrue;
        }
    }
    
    return EFalse;
}

TSmlSyncType CSyncEngine::getSyncTypeFromText(const StringBuffer& text)
{
    if (text == SYNC_ONE_WAY_FROM_CLIENT_TEXT) {
        return ESmlOneWayFromClient;
    }
    else if (text == SYNC_ONE_WAY_FROM_SERVER_TEXT) {
        return ESmlOneWayFromServer;
    }
    else if (text == SYNC_SLOW_TEXT) {
        return ESmlSlowSync;
    }
    else {
        return ESmlTwoWay;  // default
    }
}


void CSyncEngine::RunL()
{
    //Funambol::LOG.debug("RunL - status = %d, state = %d", iStatus.Int(), iAsyncAction);
    if(iStatus == KErrNone) 
    {
        switch(iAsyncAction)
        {
            case EStartSync:
            {
                StartSync(iRequestedSourceList, ESyncTriggeredAsyncAction);
                iRequestedSourceList.Reset();
                break;
            }
            case EStopSync:
            {
                StopSync();
                break;
            }
            case EEnableScheduler:
            {
                // Currently not used.
                // Can be used to trigger a scheduled sync, from a thread
                // different than the main thread.
                iScheduler->enable();
                break;
            }
            default:
                break;
        }
        iAsyncAction = EIdle;
    }
    else {
        Funambol::LOG.error("CSyncEngine::RunL() error: status = %d", iStatus.Int());
    }
}

void CSyncEngine::DoCancel()
{
    // nothing to do
}

TInt CSyncEngine::RunError(TInt aError) {
    Funambol::LOG.error("CSyncEngine RunL() error: %d", aError);
    return 0;
}


void CSyncEngine::RequestAsyncAction(const TAsyncAction aAction)
{
    //Funambol::LOG.debug("Requesting async action! state is %d", aAction);
    
    // Active object will switch to the correct method to call, based on this value.
    iAsyncAction = aAction;
    
    iStatus = KRequestPending;
    if(!IsActive()) { 
        SetActive(); 
    }
    
    // Signals that the asynchronous request associated with 'iStatus' is complete
    // Note: we must do it in the same thread that originated the request,
    //       so we signal it in the main thread.
    RThread mainThread;
    mainThread.Open(iMainThreadID);

    TRequestStatus* status = &iStatus;
    mainThread.RequestComplete(status, KErrNone);
}

TInt CSyncEngine::syncEmail()
{
    TInt err = KErrNone;
    LOG.debug("Beginning of sync email");
    logMemInfo();

    if (emailEngine == NULL) {
        emailEngine = CImapEmailEngine::NewL(this, iAppUI, SOURCE_MAIL);
    } else {
        // If another sync is already in progress, then we do not
        // start a new one
        if (emailEngine->CheckSyncInProgress()) {
            LOG.info("Email sync already in progress");
            return 1;
        }
    }

    // TODO: check if account exist?
    
    LOG.info("**** Starting sync of Emails ****");
    TRAP(err, emailEngine->PrepareL());
    if (err) {
        if (err = KErrNotFound) {
            LOG.error("Cannot sync emails: IMAP account not found");
            iAppUI->DisplayMessage(ERR_ACCOUNT_NOT_FOUND, EErrorLevel);
        }
        else {
            LOG.error("Failed to prepare Email sync: code %d", err);
        }
        return err;
    }
    TRAP(err, emailEngine->RemoteFetchL());
    if (err) {
        LOG.error("Failed to start Email sync: code %d", err);
        return err;
    }
    
    return 0;
}

void CSyncEngine::EmailSyncEnded(TCurrentSyncStatus /* status */)
{
    CImapEmailEngine* tmpEngine = emailEngine;
    emailEngine = NULL;
    delete tmpEngine;

    LOG.debug("End of sync email");
    logMemInfo();

    // Decrement the number of sources to sync
    --iNumSources;
    checkForUpdates();
}

void CSyncEngine::checkForUpdates() {
    // Once the last source ends syncing, we check for updates
    // At the end of a sync, we check if there are updates available
    if (iNumSources == 0) {
        UpdateManager updateManager(iAppUI);
        updateManager.checkForUpdates();
    }
}

void CSyncEngine::syncConfig()
{
    // We don't want more than 1 sync instance
    if (SyncActiveObject::CheckSyncInProgress()) {
        LOG.info("Config sync already in progress");
        return;
    }
    
    // Initialize the Cache File Name
    StringBuffer fullPath(PlatformAdapter::getHomeFolder());
    StringBuffer cacheFile;
    cacheFile.sprintf("%s%s", fullPath.c_str(), "config_sync_source.dat");

    TRAPD(err, SyncActiveObject::syncConfigLD(this, cacheFile));
}

void CSyncEngine::syncConfigEnded(TInt status)
{
    // Update the App UI, reporting the error if necessary
    iAppUI->UpdateSourceStatus(SOURCE_CONFIG, ESyncStatusCompleted, status);
    iAppUI->ConfigSyncEnded();
}



void CSyncEngine::enableScheduler(const bool aIsRepetitive, const unsigned int aInterval) 
{
    if (iScheduler) {
        iScheduler->setUseActiveSourceList(true);
        iScheduler->enable(aIsRepetitive, aInterval);
    }
}

void CSyncEngine::disableScheduler()
{
    if (iScheduler) {
        iScheduler->disable();
    }
}

bool CSyncEngine::isSchedulerEnabled()
{
    if (iScheduler) {
        return iScheduler->isEnabled();
    }
    return false;
}

bool CSyncEngine::removeFromScheduler(const TInt aSourceID)
{
    if (iScheduler) {
        return iScheduler->removeFromSourceList(aSourceID);
    }
    return false;
}

RArray<TInt> CSyncEngine::GetActiveSourcesList() { 
    return iAppUI->GetActiveSourcesList(); 
}
