/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2008 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */


#include "base/util/symbianUtils.h"
#include "base/util/stringUtils.h"
#include "base/util/timeUtils.h"
#include "base/SymbianLog.h"
#include "utils.h"

#include "SyncScheduler.h"


using namespace Funambol;


CSyncScheduler::CSyncScheduler() : CActive(CActive::EPriorityStandard) {}

CSyncScheduler* CSyncScheduler::NewL(CSyncEngine* aSyncEngine)
{
    CSyncScheduler* self = CSyncScheduler::NewLC(aSyncEngine);
    CleanupStack::Pop(self);
    return self;
}

CSyncScheduler* CSyncScheduler::NewLC(CSyncEngine* aSyncEngine)
{
    CSyncScheduler* self = new(ELeave) CSyncScheduler();
    self->ConstructL(aSyncEngine);
    CleanupStack::PushL(self);
    return self;
}


void CSyncScheduler::ConstructL(CSyncEngine* aSyncEngine)
{ 
    iSyncEngine = aSyncEngine;
    
    enabled    = false;
    repetitive = false;
    interval   = SCHEDULER_DEFAULT_INTERVAL;
    iSourceList.Reset();
    useActiveSourceList = false;
    
    // Initialize to now.
    iNextSchedulingTime.HomeTime();
    
    // Adds this object to the Active Scheduler
    CActiveScheduler::Add(this);
    
    // Initialize the timer
    iTimer.CreateLocal();
}


CSyncScheduler::~CSyncScheduler() 
{
    Cancel();
    iTimer.Close();
    iSourceList.Close();
}



void CSyncScheduler::enable() 
{
    enabled = true;

    TTime now;
    now.HomeTime();
    TTime newSchedulingTime(now + TTimeIntervalSeconds(interval));
    
    if ( (iNextSchedulingTime >= now) && (iNextSchedulingTime <= newSchedulingTime) ) 
    {
        // We already have another schedule in the interval window.
        // Cancel it before setting a new request.
        LOG.debug("Canceling outstanding scheduled sync");
        
        // The outstanding timer event will complete with KErrCancel.
        iTimer.Cancel();
        if(!IsActive()) { SetActive(); }
        
    }
    else {
        // No outstanding scheduled sync.
        // Schedule the next sync after 'interval' seconds from now.
        reschedule();
    }
}

void CSyncScheduler::enable(const bool aIsRepetitive) 
{
    repetitive = aIsRepetitive;
    enable();
}

void CSyncScheduler::enable(const bool aIsRepetitive, const unsigned int aInterval)
{
    repetitive = aIsRepetitive;
    interval   = aInterval;
    enable();
}

void CSyncScheduler::disable()
{
    LOG.info("Disabling the sync scheduler: outstanding requests will be discarded.");
    enabled = false;
    Cancel();           // Cancels the wait for completion of an outstanding request
    iTimer.Cancel();
    
    // Reset the next scheduling time
    iNextSchedulingTime.HomeTime();
}


void CSyncScheduler::reschedule()
{
    LOG.debug("Scheduling a new sync...");
    TTime now;
    now.HomeTime();
    iNextSchedulingTime = now + TTimeIntervalSeconds(interval);

    // Will requests an event for the active object, at that time.
    // If no errors, iStatus will be = KErrNone on completion.
    TRAPD(err, iTimer.At(iStatus, iNextSchedulingTime);)
    if (err) {
        LOG.error("Error (%d) scheduling a sync at '%s'", err, formatNextSchedulingTime().c_str());
        return;
    }
    
    if(!IsActive()) { 
        SetActive(); 
    }
    LOG.info("Next sync scheduled at '%s'", formatNextSchedulingTime().c_str());
}



void CSyncScheduler::setSourceList(const RArray<TInt>& aSourceList, const bool merge)
{
    // If we are setting a sourceList, we don't want to automatically use the
    // active source list everytime. For CP, we set the source list like this.
    useActiveSourceList = false;
    
    iCriticalSourceList.Wait();
    
    if (!merge) {
        iSourceList.Reset();
    }    
    mergeRArray(iSourceList, aSourceList);
    
    iCriticalSourceList.Signal();
    LOG.debug("exit setSourceList (tot = %d)", iSourceList.Count());
}


bool CSyncScheduler::removeFromSourceList(const TInt aSourceID)
{
    bool ret = false;
    TInt index = iSourceList.Find(aSourceID);
    
    if (index != KErrNotFound) {
        LOG.debug("Removing source '%s' from scheduler queue.", getSourceNameFromID(aSourceID).c_str());
        iCriticalSourceList.Wait();
        iSourceList.Remove(index);
        iCriticalSourceList.Signal();
        LOG.debug("%d queued sources left", iSourceList.Count());
        ret = true;
        
        if (iSourceList.Count() == 0) {
            //LOG.debug("No more sources queued: disable the scheduler");
            disable();
        }
    }

    //LOG.debug("exit removeFromSourceList (tot = %d)", iSourceList.Count());
    return ret;
}



void CSyncScheduler::clearSourceList()
{
    LOG.debug("CSyncScheduler::clearSourceList");
    iCriticalSourceList.Wait();
    iSourceList.Reset();
    iCriticalSourceList.Signal();
    
    LOG.debug("No more sources queued: disable the scheduler");
    disable();
}

StringBuffer CSyncScheduler::formatNextSchedulingTime()
{
    StringBuffer ret("");
    TBuf<50> formattedTime;
    _LIT(KFormatDateAndTime, "%*D%*N%Y%1-%2-%3 %:0%J%:1%T%:2%S");
    
    iNextSchedulingTime.FormatL(formattedTime, KFormatDateAndTime);
    ret = bufToStringBuffer(formattedTime);
    return ret;
}

unsigned int CSyncScheduler::getSecondsToNextSchedulingTime()
{
    TTimeIntervalSeconds seconds;
    TTime now;
    now.HomeTime();
    iNextSchedulingTime.SecondsFrom(now, seconds);
    
    TInt ret = seconds.Int();
    if (ret > 0) {
        return ret;
    }
    return 0;
}


void CSyncScheduler::RunL()
{
    LOG.debug("CSyncScheduler::RunL - iStatus = %d", iStatus.Int());
    if (!enabled) {
        LOG.debug("scheduler not enabled -> exit");
        return;
    }

    if(iStatus == KErrNone) 
    {
        if (useActiveSourceList) {
            // Get the list of sources that are active, now!
            iCriticalSourceList.Wait();
            iSourceList = iSyncEngine->GetActiveSourcesList();
            iCriticalSourceList.Signal();
            LOG.debug("SyncScheduler reads active sourceList: tot = %d", iSourceList.Count());
        }

        if (iSourceList.Count()) {
            //
            // Start the sync!
            //
            LOG.debug("Starting scheduled sync (%d sources)", iSourceList.Count());
            iSyncEngine->StartSync(iSourceList, ESyncTriggeredScheduled);
        }
        else {
            // It's not necessary to reschedule a new sync if iSourceList empty.
            // - For CP: the scheduler will be re-enabled at the next modification.
            // - For schedule: we restart the service when the first source is enabled.
            LOG.debug("Don't start the scheduled sync: no sources to schedule.");
            disable();
            return;
        }
        
        if (repetitive) {
            // Reschedule, and don't clear the list of source ID
            reschedule();
        }
        else {
            // Not repetitive: clear the list of source ID
            clearSourceList();
        }
    }
    
    else if (iStatus == KErrCancel)
    {
        //
        // Outstanding request was canceled.
        //
        if (repetitive) {
            // Reschedule, and don't clear the list of source ID
            reschedule();
        }
        else {
            // Not repetitive: clear the list of source ID
            clearSourceList();
        }
    }
    
    else if (iStatus == KErrAbort) 
    {
        // Request was aborted (???) -> Reschedule it.
        LOG.debug("Request was aborted -> rescheduling it...");

        // Will requests an event for the active object, at that time.
        // If no errors, iStatus will be = KErrNone on completion.
        TRAPD(err, iTimer.At(iStatus, iNextSchedulingTime);)
        if (err) {
            LOG.error("Error (%d) scheduling a sync at '%s'", err, formatNextSchedulingTime().c_str());
            return;
        }
        
        if(!IsActive()) { 
            SetActive(); 
        }
        LOG.info("Sync rescheduled at '%s'", formatNextSchedulingTime().c_str());
    }
    
    //LOG.debug("exiting CSyncScheduler::RunL");
}


void CSyncScheduler::DoCancel() 
{
    //LOG.debug("CSyncScheduler::DoCancel");
    iTimer.Cancel();
}

TInt CSyncScheduler::RunError(TInt aError) {
    LOG.error("CSyncScheduler RunL() error: %d", aError);
    return 0;
}





